import asyncio
import base64
import json
import os
import warnings
from io import StringIO
from typing import Any, Dict, List, Literal, Optional

import aiohttp
import pandas as pd
from aiolimiter import AsyncLimiter

from viadot.config import get_source_credentials
from viadot.exceptions import APIError, CredentialError
from viadot.signals import SKIP
from viadot.sources.base import Source
from viadot.utils import handle_api_response, validate

warnings.simplefilter("ignore")


class Genesys(Source):
    def __init__(
        self,
        view_type: str = "queue_performance_detail_view",
        ids_mapping: Dict[str, Any] = None,
        start_date: str = None,
        end_date: str = None,
        report_name: str = None,
        file_extension: Literal["xls", "xlsx", "csv"] = "csv",
        config_key: str = None,
        credentials: Dict[str, Any] = None,
        environment: str = None,
        report_url: str = None,
        schedule_id: str = None,
        report_columns: List[str] = None,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Genesys connector which allows for reports scheduling, listing and downloading into DataFrame or specified format output.

        Args:
            view_type (str, optional): The type of view export job to be created. Defaults to "queue_performance_detail_view".
            ids_mapping (str, optional): Dictionary mapping for converting IDs to strings. Defaults to None.
            start_date (str, optional):  Start date of the report. Defaults to None.
            end_date (str, optional):  End date of the report. Defaults to None.
            report_name (str, optional): Name of the report. Defaults to None.
            file_extension (Literal[xls, xlsx, csv], optional): File extensions for downloaded files. Defaults to "csv".
            credentials (Dict[str, Any], optional): Credentials to connect with Genesys API containing CLIENT_ID,
            environment (str, optional): Adress of host server. Defaults to None than will be used enviroment
            from credentials.
            report_url (str, optional): The url of report generated in json response. Defaults to None.
            schedule_id (str, optional): The ID of report. Defaults to None.
            report_columns (List[str], optional): List of exisiting column in report. Defaults to None.

        Raises:
            CredentialError: If credentials are not provided in local_config or directly as a parameter.
        """

        credentials = credentials or get_source_credentials(config_key) or {}

        if credentials is None:
            raise CredentialError("Please specify the credentials.")

        self.credentials = credentials

        super().__init__(*args, credentials=self.credentials, **kwargs)

        self.view_type = view_type
        self.schedule_id = schedule_id
        self.report_name = report_name
        self.environment = environment
        self.report_url = report_url
        self.report_columns = report_columns

        self.start_date = start_date
        self.end_date = end_date
        self.file_extension = file_extension
        self.ids_mapping = ids_mapping
        self.count = iter(range(99999))

        if self.schedule_id is None:
            self.schedule_id = self.credentials.get("SCHEDULE_ID", None)

        if self.environment is None:
            self.environment = self.credentials.get("ENVIRONMENT", None)

        if self.ids_mapping is None:
            self.ids_mapping = self.credentials.get("IDS_MAPPING", None)

            if isinstance(self.ids_mapping, dict) and self.ids_mapping is not None:
                self.logger.info("IDS_MAPPING loaded from local credential.")
            else:
                self.logger.warning(
                    "IDS_MAPPING is not provided in you credentials or is not a dictionary."
                )

        self.report_data = []

    @property
    def authorization_token(self, verbose: bool = False) -> Dict[str, Any]:
        """
        Get authorization token with request headers.

        Args:
            verbose (bool, optional): Switch on/off for logging messages. Defaults to False.

        Returns:
            Dict[str, Any]: Request headers with token.
        """
        CLIENT_ID = self.credentials.get("CLIENT_ID", None)
        CLIENT_SECRET = self.credentials.get("CLIENT_SECRET", None)
        authorization = base64.b64encode(
            bytes(CLIENT_ID + ":" + CLIENT_SECRET, "ISO-8859-1")
        ).decode("ascii")
        request_headers = {
            "Authorization": f"Basic {authorization}",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        request_body = {"grant_type": "client_credentials"}
        response = handle_api_response(
            f"https://login.{self.environment}/oauth/token",
            data=request_body,
            headers=request_headers,
            method="POST",
            timeout=3600,
        )
        if verbose:
            if response.status_code == 200:
                self.logger.info("Temporary authorization token was generated.")
            else:
                self.logger.info(
                    f"Failure: { str(response.status_code) } - { response.reason }"
                )
        response_json = response.json()
        request_headers = {
            "Authorization": f"{ response_json['token_type'] } { response_json['access_token']}",
            "Content-Type": "application/json",
        }

        return request_headers

    def genesys_generate_exports(
        self, post_data_list: List[str], end_point: str = "reporting/exports"
    ) -> Optional[dict]:
        """Function that make POST request method to generate export reports.

        Args:
            post_data_list (List[str], optional): List of string templates to generate json body. Defaults to None.
            end_point (str, optional): Final end point for Genesys connection. Defaults to "reporting/exports".

        Returns:
            Optional[dict]: Dict when the "conversations" endpoint is called, otherwise returns None.
        """

        limiter = AsyncLimiter(2, 15)
        semaphore = asyncio.Semaphore(value=1)

        async def generate_post():
            cnt = 0

            for data_to_post in post_data_list:
                if cnt < 10:
                    payload = json.dumps(data_to_post)
                    async with aiohttp.ClientSession() as session:
                        await semaphore.acquire()
                        async with limiter:
                            async with session.post(
                                f"https://api.{self.environment}/api/v2/analytics/{end_point}",
                                headers=self.authorization_token,
                                data=payload,
                            ) as resp:
                                global new_report
                                new_report = await resp.read()
                                self.logger.info(
                                    f"Generated report export --- \n {payload}."
                                )
                                semaphore.release()
                    cnt += 1
                else:
                    await asyncio.sleep(3)
                    cnt = 0

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError as e:
            if str(e).startswith("There is no current event loop in thread"):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            else:
                raise e

        coroutine = generate_post()
        loop.run_until_complete(coroutine)

        if end_point == "conversations/details/query":
            return json.loads(new_report.decode("utf-8"))

    def load_reporting_exports(
        self, page_size: int = 100, verbose: bool = False
    ) -> Dict[str, Any]:
        """
        GET method for reporting export.

        Args:
            page_size (int, optional): The number of items on page to print. Defaults to 100.
            verbose (bool, optional): Switch on/off for logging messages. Defaults to False.

        Returns:
            Dict[str, Any]: schedule genesys report.
        """
        new_report = handle_api_response(
            url=f"https://api.{self.environment}/api/v2/analytics/reporting/exports?pageSize={page_size}",
            headers=self.authorization_token,
            method="GET",
        )

        if new_report.status_code == 200:
            if verbose:
                self.logger.info("Succesfully loaded all exports.")
            return new_report.json()
        else:
            self.logger.error(f"Failed to loaded all exports. - {new_report.content}")
            raise APIError("Failed to loaded all exports.")

    def get_reporting_exports_data(self) -> None:
        """
        Function that generate list of reports metadata for further processing steps.

        Returns:
            None
        """
        request_json = self.load_reporting_exports()

        if request_json is not None:
            entities = request_json.get("entities")
            assert isinstance(entities, list)
            if len(entities) != 0:
                for entity in entities:
                    tmp = [
                        entity.get("id"),
                        entity.get("downloadUrl"),
                        entity.get("filter").get("queueIds", [-1])[0],
                        entity.get("filter").get("mediaTypes", [-1])[0],
                        entity.get("viewType"),
                        entity.get("interval"),
                        entity.get("status"),
                    ]
                    self.report_data.append(tmp)
            assert len(self.report_data) > 0
        self.logger.info("Generated list of reports entities.")

    def download_report(
        self,
        report_url: str,
        output_file_name: str = None,
        file_extension: str = "csv",
        path: str = "",
        sep: str = "\t",
        drop_duplicates: bool = True,
    ) -> None:
        """
        Download report to excel file.

        Args:
            report_url (str): url to report, fetched from json response.
            output_file_name (str, optional): Output file name. Defaults to None.
            file_extension (str, optional): Output file extension. Defaults to "xls".
            path (str, optional): Path to the generated excel file. Defaults to empty string.
            sep (str, optional): Separator in csv file. Defaults to "\t".
            drop_duplicates (bool, optional): Decide if drop duplicates. Defaults to True.

        Returns:
            None
        """
        response_file = handle_api_response(
            url=f"{report_url}", headers=self.authorization_token
        )
        if output_file_name is None:
            final_file_name = f"Genesys_Queue_Metrics_Interval_Export.{file_extension}"
        else:
            final_file_name = f"{output_file_name}.{file_extension}"

        df = pd.read_csv(StringIO(response_file.content.decode("utf-8")))

        if drop_duplicates is True:
            df.drop_duplicates(inplace=True, ignore_index=True)

        df.to_csv(os.path.join(path, final_file_name), index=False, sep=sep)

    def download_all_reporting_exports(
        self, store_file_names: bool = True, path: str = ""
    ) -> List[str]:
        """
        Get information form data report and download all files.

        Args:
            store_file_names (bool, optional): decide whether to store list of names. Defaults to True.
            path (str, optional): Path to the generated excel file. Defaults to empty string.

        Returns:
            List[str]: All file names of downloaded files.
        """
        file_name_list = []
        temp_ids_mapping = self.ids_mapping
        if temp_ids_mapping is None:
            self.logger.warning("IDS_MAPPING is not provided in you credentials.")
        else:
            self.logger.info("IDS_MAPPING loaded from local credential.")

        for single_report in self.report_data:
            self.logger.info(single_report)
            if single_report[-1] == "RUNNING":
                self.logger.warning(
                    "The request is still in progress and will be deleted, consider add more seconds in `view_type_time_sleep` parameter."
                )
                continue
            elif single_report[-1] == "FAILED":
                self.logger.warning(
                    "This message 'FAILED_GETTING_DATA_FROM_SERVICE' raised during script execution."
                )
                continue
            elif self.start_date not in single_report[5]:
                self.logger.warning(
                    f"The report with ID {single_report[0]} doesn't match with the interval date that you have already defined. \
                        The report won't be downloaded but will be deleted."
                )
                continue

            if single_report[4].lower() == "queue_performance_detail_view":
                file_name = (
                    temp_ids_mapping.get(single_report[2]) + "_" + single_report[3]
                ).upper()
            elif single_report[4].lower() in [
                "agent_performance_summary_view",
                "agent_status_summary_view",
            ]:
                date = self.start_date.replace("-", "")
                file_name = self.view_type.upper() + "_" + f"{date}"
            elif single_report[4].lower() in [
                "agent_status_detail_view",
            ]:
                date = self.start_date.replace("-", "")
                file_name = self.view_type.upper() + f"_{next(self.count)}_" + f"{date}"
            else:
                raise SKIP(
                    message=f"View type {self.view_type} not defined in viadot, yet..."
                )

            self.download_report(
                report_url=single_report[1],
                path=path,
                output_file_name=file_name,
                file_extension=self.file_extension,
            )
            if store_file_names is True:
                file_name_list.append(file_name + "." + self.file_extension)

        self.logger.info("Al reports were successfully dowonload.")

        if store_file_names is True:
            self.logger.info("Successfully genetared file names list.")
            return file_name_list

    def generate_reporting_export(
        self, data_to_post: Dict[str, Any], verbose: bool = False
    ) -> int:
        """
        POST method for reporting export.

        Args:
            data_to_post (Dict[str, Any]): Json format of POST body.
            verbose (bool, optional): Decide if enable logging. Defaults to True.

        Returns:
            int: Status code.
        """
        payload = json.dumps(data_to_post)
        new_report = handle_api_response(
            url=f"https://api.{self.environment}/api/v2/analytics/reporting/exports",
            headers=self.authorization_token,
            method="POST",
            data=payload,
        )
        if verbose:
            if new_report.status_code == 200:
                self.logger.info("Succesfully generated new export.")
            else:
                self.logger.error(
                    f"Failed to generated new export. - {new_report.content}"
                )
                raise APIError("Failed to generated new export.")
        return new_report.status_code

    def delete_reporting_exports(self, report_id: str) -> int:
        """DELETE method for deleting particular reporting exports.

        Args:
            report_id (str): Defined at the end of report url.

        Returns:
            int: Status code.
        """
        delete_method = handle_api_response(
            url=f"https://api.{self.environment}/api/v2/analytics/reporting/exports/{report_id}",
            headers=self.authorization_token,
            method="DELETE",
        )
        if delete_method.status_code < 300:
            self.logger.info("Successfully deleted report from Genesys API.")

        else:
            self.logger.error(
                f"Failed to deleted report from Genesys API. - {delete_method.content}"
            )
            raise APIError("Failed to deleted report from Genesys API.")

        return delete_method.status_code

    def delete_all_reporting_exports(self) -> None:
        """
        Function that deletes all reporting from self.reporting_data list.

        Returns:
            None
        """
        for report in self.report_data:
            status_code = self.delete_reporting_exports(report_id=report[0])
            assert status_code < 300

        self.logger.info("Successfully removed all reports.")

    def get_analitics_url_report(self) -> str:
        """
        Fetching analytics report url from json response.

        Returns:
            str: Url for analytics report
        """
        response = handle_api_response(
            url=f"https://api.{self.environment}/api/v2/analytics/reporting/schedules/{self.schedule_id}",
            headers=self.authorization_token,
        )
        try:
            response_json = response.json()
            report_url = response_json.get("lastRun", None).get("reportUrl", None)
            self.logger.info("Successfully downloaded report from genesys api")
            return report_url
        except AttributeError as e:
            self.logger.error(
                "Output data error: " + str(type(e).__name__) + ": " + str(e)
            )

    def get_all_schedules_job(self) -> Dict[str, Any]:
        """
        Fetching analytics report url from json response.

        Returns:
            Dict[str, Any]: Json body with all schedules jobs.
        """
        response = handle_api_response(
            url=f"https://api.{self.environment}/api/v2/analytics/reporting/schedules",
            headers=self.authorization_token,
        )
        try:
            response_json = response.json()
            self.logger.info("Successfully downloaded schedules jobs.")
            return response_json
        except AttributeError as e:
            self.logger.error(
                "Output data error: " + str(type(e).__name__) + ": " + str(e)
            )

    def schedule_report(self, data_to_post: Dict[str, Any]) -> int:
        """
        POST method for report scheduling.

        Args:
            data_to_post (Dict[str, Any]): Json format of POST body.

        Returns:
            int: status code
        """
        payload = json.dumps(data_to_post)
        new_report = handle_api_response(
            url=f"https://api.{self.environment}/api/v2/analytics/reporting/schedules",
            headers=self.authorization_token,
            method="POST",
            data=payload,
        )
        if new_report.status_code == 200:
            self.logger.info("Succesfully scheduled new report.")
        else:
            self.logger.error(f"Failed to scheduled new report. - {new_report.content}")
            raise APIError("Failed to scheduled new report.")
        return new_report.status_code

    def to_df(
        self,
        report_url: str = None,
        tests: dict = None,
    ) -> pd.DataFrame:
        """
        Download genesys data into a pandas DataFrame.

        Args:
            report_url (str): Report url from api response.
            tests (Dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.

        Returns:
            pd.DataFrame: The DataFrame with time range.
        """
        if report_url is None:
            report_url = self.get_analitics_url_report()
        response_file = handle_api_response(
            url=f"{report_url}", headers=self.authorization_token
        )
        if self.report_columns is None:
            df = pd.read_excel(response_file.content, header=6)
        else:
            df = pd.read_excel(
                response_file.content, names=self.report_columns, skiprows=6
            )

        if tests:
            validate(df=df, tests=tests)

        return df

    def delete_scheduled_report_job(self, report_id: str) -> int:
        """
        DELETE method for deleting particular report job.

        Args:
            report_id (str): Defined at the end of report url.

        Returns:
            int: Status code.
        """
        delete_method = handle_api_response(
            url=f"https://api.{self.environment}/api/v2/analytics/reporting/schedules/{report_id}",
            headers=self.authorization_token,
            method="DELETE",
        )
        if delete_method.status_code == 200:
            self.logger.info("Successfully deleted report from Genesys API.")

        else:
            self.logger.error(
                f"Failed to deleted report from Genesys API. - {delete_method.content}"
            )
            raise APIError("Failed to deleted report from Genesys API.")

        return delete_method.status_code



import copy
import json
import os
import shutil
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import List, Literal, Union, cast

import pandas as pd
import prefect
import pyarrow as pa
import pyarrow.dataset as ds
from prefect import Flow, Task, task
from prefect.engine.state import Failed
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities import logging
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail
from toolz import curry
from visions.functional import infer_type
from visions.typesets.complete_set import CompleteSet

from viadot.config import local_config
from viadot.tasks import AzureKeyVaultSecret

logger = logging.get_logger()
METADATA_COLUMNS = {"_viadot_downloaded_at_utc": "DATETIME"}


@task
def add_ingestion_metadata_task(
    df: pd.DataFrame,
):
    """Add ingestion metadata columns, eg. data download date

    Args:
        df (pd.DataFrame): input DataFrame.
    """
    df2 = df.copy(deep=True)
    df2["_viadot_downloaded_at_utc"] = datetime.now(timezone.utc).replace(microsecond=0)
    return df2


@task
def get_latest_timestamp_file_path(files: List[str]) -> str:
    """
    Return the name of the latest file in a given data lake directory,
    given a list of paths in that directory. Such list can be obtained using the
    `AzureDataLakeList` task. This task is useful for working with immutable data lakes as
    the data is often written in the format /path/table_name/TIMESTAMP.parquet.
    """

    logger = prefect.context.get("logger")

    file_names = [Path(file).stem for file in files]
    latest_file_name = max(file_names, key=lambda d: datetime.fromisoformat(d))
    latest_file = files[file_names.index(latest_file_name)]

    logger.debug(f"Latest file: {latest_file}")

    return latest_file


@task
def dtypes_to_json_task(dtypes_dict, local_json_path: str):
    """
    Creates json file from a dictionary.
    Args:
        dtypes_dict (dict): Dictionary containing data types.
        local_json_path (str): Path to local json file.
    """
    with open(local_json_path, "w") as fp:
        json.dump(dtypes_dict, fp)


@task
def chunk_df(df: pd.DataFrame, size: int = 10_000) -> List[pd.DataFrame]:
    """
    Creates pandas Dataframes list of chunks with a given size.
    Args:
        df (pd.DataFrame): Input pandas DataFrame.
        size (int, optional): Size of a chunk. Defaults to 10000.
    """
    n_rows = df.shape[0]
    chunks = [df[i : i + size] for i in range(0, n_rows, size)]
    return chunks


@task
def df_get_data_types_task(df: pd.DataFrame) -> dict:
    """
    Returns dictionary containing datatypes of pandas DataFrame columns.
    Args:
        df (pd.DataFrame): Input pandas DataFrame.
    """
    typeset = CompleteSet()
    dtypes = infer_type(df, typeset)
    dtypes_dict = {k: str(v) for k, v in dtypes.items()}
    return dtypes_dict


@task
def get_sql_dtypes_from_df(df: pd.DataFrame) -> dict:
    """Obtain SQL data types from a pandas DataFrame"""
    typeset = CompleteSet()
    dtypes = infer_type(df.head(10000), typeset)
    dtypes_dict = {k: str(v) for k, v in dtypes.items()}
    dict_mapping = {
        "Float": "REAL",
        "Image": None,
        "Categorical": "VARCHAR(500)",
        "Time": "TIME",
        "Boolean": "VARCHAR(5)",  # Bool is True/False, Microsoft expects 0/1
        "DateTime": "DATETIMEOFFSET",  # DATETIMEOFFSET is the only timezone-aware dtype in TSQL
        "Object": "VARCHAR(500)",
        "EmailAddress": "VARCHAR(50)",
        "File": None,
        "Geometry": "GEOMETRY",
        "Ordinal": "INT",
        "Integer": "INT",
        "Generic": "VARCHAR(500)",
        "UUID": "VARCHAR(50)",  # Microsoft uses a custom UUID format so we can't use it
        "Complex": None,
        "Date": "DATE",
        "String": "VARCHAR(500)",
        "IPAddress": "VARCHAR(39)",
        "Path": "VARCHAR(255)",
        "TimeDelta": "VARCHAR(20)",  # datetime.datetime.timedelta; eg. '1 days 11:00:00'
        "URL": "VARCHAR(255)",
        "Count": "INT",
    }
    dict_dtypes_mapped = {}
    for k in dtypes_dict:
        dict_dtypes_mapped[k] = dict_mapping[dtypes_dict[k]]

    # This is required as pandas cannot handle mixed dtypes in Object columns
    dtypes_dict_fixed = {
        k: ("String" if v == "Object" else str(v))
        for k, v in dict_dtypes_mapped.items()
    }

    return dtypes_dict_fixed


@task
def update_dict(d: dict, d_new: dict) -> dict:
    d_copy = copy.deepcopy(d)
    d_copy.update(d_new)
    return d_copy


@task
def df_map_mixed_dtypes_for_parquet(
    df: pd.DataFrame, dtypes_dict: dict
) -> pd.DataFrame:
    """
    Pandas is not able to handle mixed dtypes in the column in to_parquet
    Mapping 'object' visions dtype to 'string' dtype to allow Pandas to_parquet

    Args:
        dict_dtypes_mapped (dict): Data types dictionary inferenced by Visions
        df (pd.DataFrame): input DataFrame.

    Returns:
        df_mapped (pd.DataFrame): Pandas DataFrame with mapped Data Types to workaround Pandas to_parquet bug connected with mixed dtypes in object:.
    """
    df_mapped = df.copy()
    for col, dtype in dtypes_dict.items():
        if dtype == "Object":
            df_mapped[col] = df_mapped[col].astype("string")
    return df_mapped


@task
def update_dtypes_dict(dtypes_dict: dict) -> dict:
    """
    Task to update dtypes_dictionary that will be stored in the schema. It's required due to workaround Pandas to_parquet bug connected with mixed dtypes in object

    Args:
        dtypes_dict (dict): Data types dictionary inferenced by Visions

    Returns:
        dtypes_dict_updated (dict): Data types dictionary updated to follow Pandas requeirments in to_parquet functionality.
    """
    dtypes_dict_updated = {
        k: ("String" if v == "Object" else str(v)) for k, v in dtypes_dict.items()
    }

    return dtypes_dict_updated


@task
def df_to_csv(
    df: pd.DataFrame,
    path: str,
    sep="\t",
    if_exists: Literal["append", "replace", "skip"] = "replace",
    **kwargs,
) -> None:
    """
    Task to create csv file based on pandas DataFrame.
    Args:
    df (pd.DataFrame): Input pandas DataFrame.
    path (str): Path to output csv file.
    sep (str, optional): The separator to use in the CSV. Defaults to "\t".
    if_exists (Literal["append", "replace", "skip"], optional): What to do if the table exists. Defaults to "replace".
    """

    if if_exists == "append" and os.path.isfile(path):
        csv_df = pd.read_csv(path, sep=sep)
        out_df = pd.concat([csv_df, df])
    elif if_exists == "replace":
        out_df = df
    elif if_exists == "skip" and os.path.isfile(path):
        logger.info("Skipped.")
        return
    else:
        out_df = df

    # create directories if they don't exist
    try:
        if not os.path.isfile(path):
            directory = os.path.dirname(path)
            os.makedirs(directory, exist_ok=True)
    except Exception:
        pass

    out_df.to_csv(path, index=False, sep=sep)


@task
def df_to_parquet(
    df: pd.DataFrame,
    path: str,
    if_exists: Literal["append", "replace", "skip"] = "replace",
    **kwargs,
) -> None:
    """
    Task to create parquet file based on pandas DataFrame.
    Args:
    df (pd.DataFrame): Input pandas DataFrame.
    path (str): Path to output parquet file.
    if_exists (Literal["append", "replace", "skip"], optional): What to do if the table exists. Defaults to "replace".
    """
    if if_exists == "append" and os.path.isfile(path):
        parquet_df = pd.read_parquet(path)
        out_df = pd.concat([parquet_df, df])
    elif if_exists == "replace":
        out_df = df
    elif if_exists == "skip":
        logger.info("Skipped.")
        return
    else:
        out_df = df

    # create directories if they don't exist
    try:
        if not os.path.isfile(path):
            directory = os.path.dirname(path)
            os.makedirs(directory, exist_ok=True)
    except Exception:
        pass

    out_df.to_parquet(path, index=False, **kwargs)


@task
def union_dfs_task(dfs: List[pd.DataFrame]):
    """
    Create one DataFrame from a list of pandas DataFrames.
    Args:
        dfs (List[pd.DataFrame]): List of pandas Dataframes to concat. In case of different size of DataFrames NaN values can appear.
    """
    return pd.concat(dfs, ignore_index=True)


@task
def write_to_json(dict_, path):
    """
    Creates json file from a dictionary. Log record informs about the writing file proccess.
    Args:
        dict_ (dict): Dictionary.
        path (str): Path to local json file.
    """
    logger = prefect.context.get("logger")

    if os.path.isfile(path):
        logger.warning(f"File {path} already exists. Overwriting...")
    else:
        logger.debug(f"Writing to {path}...")

    # create parent directories if they don't exist
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    with open(path, mode="w") as f:
        json.dump(dict_, f)

    logger.debug(f"Successfully wrote to {path}.")


@task
def cleanup_validation_clutter(expectations_path):
    ge_project_path = Path(expectations_path).parent
    shutil.rmtree(ge_project_path)


@task
def df_converts_bytes_to_int(df: pd.DataFrame) -> pd.DataFrame:
    logger = prefect.context.get("logger")
    logger.info("Converting bytes in dataframe columns to list of integers")
    return df.applymap(lambda x: list(map(int, x)) if isinstance(x, bytes) else x)  # noqa


@task(
    max_retries=3,
    retry_delay=timedelta(seconds=10),
)
def df_to_dataset(
    df: pd.DataFrame, partitioning_flavor="hive", format="parquet", **kwargs
) -> None:
    """
    Use `pyarrow.dataset.write_to_dataset()` to write from a pandas DataFrame to a dataset.
    This enables several data lake-specific optimizations such as parallel writes, partitioning,
    and file size (via `max_rows_per_file` parameter).

    Args:
        df (pd.DataFrame): The pandas DataFrame to write.
        partitioning_flavor (str, optional): The partitioning flavor to use. Defaults to "hive".
        format (str, optional): The dataset format. Defaults to 'parquet'.
        kwargs: Keyword arguments to be passed to `write_to_dataset()`. See
        https://arrow.apache.org/docs/python/generated/pyarrow.dataset.write_dataset.html.

    Examples:
        table = pa.Table.from_pandas(df_contact)
        base_dir = "/home/viadot/contact"
        partition_cols = ["updated_at_year", "updated_at_month", "updated_at_day"]

        df_to_dataset(
            data=table,
            base_dir=base_dir,
            partitioning=partition_cols,
            existing_data_behavior='overwrite_or_ignore',
            max_rows_per_file=100_000
        )
    """
    table = pa.Table.from_pandas(df)
    ds.write_dataset(
        data=table, partitioning_flavor=partitioning_flavor, format=format, **kwargs
    )


@curry
def custom_mail_state_handler(
    tracked_obj: Union["Flow", "Task"],
    old_state: prefect.engine.state.State,
    new_state: prefect.engine.state.State,
    only_states: list = [Failed],
    local_api_key: str = None,
    credentials_secret: str = None,
    vault_name: str = None,
    from_email: str = None,
    to_emails: str = None,
) -> prefect.engine.state.State:
    """
    Custom state handler configured to work with sendgrid.
    Works as a standalone state handler, or can be called from within a custom state handler.
    Args:
        tracked_obj (Task or Flow): Task or Flow object the handler is registered with.
        old_state (State): previous state of tracked object.
        new_state (State): new state of tracked object.
        only_states ([State], optional): similar to `ignore_states`, but instead _only_
            notifies you if the Task / Flow is in a state from the provided list of `State`
            classes.
        local_api_key (str, optional): Api key from local config.
        credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with API KEY.
        vault_name (str, optional): Name of key vault.
        from_email (str): Sender mailbox address.
        to_emails (str): Receiver mailbox address.
    Returns: State: the `new_state` object that was provided

    """

    if credentials_secret is None:
        try:
            credentials_secret = PrefectSecret("mail_notifier_api_key").run()
        except ValueError:
            pass

    if credentials_secret is not None:
        credentials_str = AzureKeyVaultSecret(
            credentials_secret, vault_name=vault_name
        ).run()
        api_key = json.loads(credentials_str).get("API_KEY")
    elif local_api_key is not None:
        api_key = local_config.get(local_api_key).get("API_KEY")
    else:
        raise Exception("Please provide API KEY")

    curr_dt = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    only_states = only_states or []
    if only_states and not any(
        [isinstance(new_state, included) for included in only_states]
    ):
        return new_state
    url = prefect.client.Client().get_cloud_url(
        "flow-run", prefect.context["flow_run_id"], as_user=False
    )
    message = Mail(
        from_email=from_email,
        to_emails=to_emails,
        subject=f"The flow {tracked_obj.name} - Status {new_state}",
        html_content=f"<strong>The flow {cast(str,tracked_obj.name)} FAILED at {curr_dt}. \
    <p>More details here: {url}</p></strong>",
    )
    try:
        sendgrid = SendGridAPIClient(api_key)
        sendgrid.send(message)
    except Exception as e:
        raise e

    return new_state


@task
def df_clean_column(
    df: pd.DataFrame, columns_to_clean: List[str] = None
) -> pd.DataFrame:
    """
    Function that removes special characters (such as escape symbols)
    from a pandas DataFrame.

    Args:
    df (pd.DataFrame): The DataFrame to clean.
    columns_to_clean (List[str]): A list of columns to clean. Defaults is None.

    Returns:
    pd.DataFrame: The cleaned DataFrame
    """

    df = df.copy()

    if columns_to_clean is None:
        df.replace(
            to_replace=[r"\\t|\\n|\\r", "\t|\n|\r"],
            value=["", ""],
            regex=True,
            inplace=True,
        )
    else:
        for col in columns_to_clean:
            df[col].replace(
                to_replace=[r"\\t|\\n|\\r", "\t|\n|\r"],
                value=["", ""],
                regex=True,
                inplace=True,
            )
    return df



from pathlib import Path
from typing import Generator, Literal

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import s3fs
import urllib3
from minio import Minio
from minio.error import S3Error
from pydantic import BaseModel
from urllib3.exceptions import NewConnectionError

from viadot.config import get_source_credentials
from viadot.sources.base import Source


class MinIOCredentials(BaseModel):
    endpoint: str = "localhost:9000"
    access_key: str
    secret_key: str
    bucket: str
    secure: bool = True  # Whether to use HTTPS.
    verify: bool = True  # Whether to verify TLS certificates.


class MinIO(Source):
    """
    A class for interacting with MinIO, in a more Pythonic, user-friendly, and robust
    way than the official minio client.

    Args:
        credentials (MinIOCredentials): MinIO credentials.
        config_key (str, optional): The key in the viadot config holding relevant credentials.
    """

    def __init__(
        self,
        credentials: MinIOCredentials = None,
        config_key: str = None,
        *args,
        **kwargs,
    ):
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = MinIOCredentials(**raw_creds).dict(
            by_alias=True
        )  # validate the credentials

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.endpoint = self.credentials.get("endpoint")
        self.access_key = self.credentials.get("access_key")
        self.secret_key = self.credentials.get("secret_key")
        self.bucket = self.credentials.get("bucket")
        self.secure = self.credentials.get("secure")
        self.verify = self.credentials.get("verify")

        self.http_scheme = "https" if self.secure is True else "http"

        self.client = Minio(
            self.endpoint,
            access_key=self.access_key,
            secret_key=self.secret_key,
            secure=self.secure,
            http_client=urllib3.PoolManager(
                timeout=urllib3.Timeout.DEFAULT_TIMEOUT,
                retries=urllib3.Retry(
                    connect=1,
                    read=3,
                    total=2,
                    backoff_factor=1,
                    status_forcelist=[500, 502, 503, 504],
                ),
                cert_reqs="CERT_REQUIRED" if self.verify else "CERT_NONE",
            ),
        )

        self.storage_options = {
            "key": self.access_key,
            "secret": self.secret_key,
            "client_kwargs": {
                "endpoint_url": f"{self.http_scheme}://" + self.endpoint,
                "use_ssl": self.secure,
                "verify": self.verify,
            },
        }

        self.fs = s3fs.S3FileSystem(
            key=self.storage_options["key"],
            secret=self.storage_options["secret"],
            client_kwargs=self.storage_options["client_kwargs"],
            anon=False,
        )

    def from_arrow(
        self,
        table: pa.Table,
        schema_name: str | None = None,
        table_name: str | None = None,
        path: str | Path | None = None,
        partition_cols: list[str] | None = None,
        if_exists: Literal["error", "delete_matching", "overwrite_or_ignore"] = "error",
    ):
        """
        Create a Parquet dataset on MinIO from a PyArrow Table.

        Uses multi-part upload to upload the table in chunks, speeding up the
        process by using multithreading and avoiding upload size limits.

        Either both `schema_name` and `table_name` or only `path` must be provided.

        `path` allows specifying an arbitrary path, while `schema_name` and `table_name`
        provide a shortcut for creating a data lakehouse-like structure of
        `s3://<bucket>/<schema_name>/<table_name>/<table_name>.parquet`.

        For more information on partitioning, see
        https://arrow.apache.org/docs/python/generated/pyarrow.parquet.write_to_dataset.html#pyarrow-parquet-write-to-dataset  # noqa

        Args:
            table (pa.Table): The table to upload.
            schema_name (str, optional): The name of the schema directory. Defaults to
                None.
            table_name (str, optional): The name of the table directory. Defaults to
                None.
            path (str | Path, optional): The path to the destination file. Defaults to
                None.
            partition_cols (list[str], optional): The columns to partition by. Defaults
                to None.
            if_exists (Literal["error", "delete_matching", "overwrite_or_ignore"],
                optional). What to do if the dataset already exists.
        """
        fqn_or_path = (schema_name and table_name) or (
            path and not (schema_name or table_name)
        )
        if not fqn_or_path:
            raise ValueError(
                "Either both `schema_name` and `table_name` or only `path` must be provided."
            )

        # We need to create the dirs here as PyArrow also tries to create the bucket,
        # which shouldn't be allowed for whomever is executing this code.
        self.logger.debug(f"Creating directory for table {table_name}...")
        path = path or f"{schema_name}/{table_name}"
        self.fs.makedirs(path, exist_ok=True)
        self.logger.debug("Directory has been created successfully.")

        # Write the data.
        pq.write_to_dataset(
            table,
            root_path=path,
            partition_cols=partition_cols,
            existing_data_behavior=if_exists,
            filesystem=self.fs,
            max_rows_per_file=1024 * 1024,
            create_dir=False,  # Required as Arrow attempts to create the bucket, too.
        )

    def from_df(
        self,
        df: pd.DataFrame,
        schema_name: str | None = None,
        table_name: str | None = None,
        path: str | Path | None = None,
        partition_cols: list[str] | None = None,
        if_exists: Literal["error", "delete_matching", "overwrite_or_ignore"] = "error",
    ) -> None:
        """
        Create a Parquet dataset on MinIO from a PyArrow Table.

        Uses multi-part upload to upload the table in chunks, speeding up the
        process by using multithreading and avoiding upload size limits.

        Either both `schema_name` and `table_name` or only `path` must be provided.

        `path` allows specifying an arbitrary path, while `schema_name` and `table_name`
        provide a shortcut for creating a data lakehouse-like structure of
        `s3://<bucket>/<schema_name>/<table_name>/<table_name>.parquet`.

        For more information on partitioning, see
        https://arrow.apache.org/docs/python/generated/pyarrow.parquet.write_to_dataset.html#pyarrow-parquet-write-to-dataset  # noqa

        Args:
            df (pd.DataFrame): The DataFrame to upload.
            schema_name (str, optional): The name of the schema directory. Defaults to
                None.
            table_name (str, optional): The name of the table directory. Defaults to
                None.
            path (str | Path, optional): The path to the destination file. Defaults to
                None.
            partition_cols (list[str], optional): The columns to partition by. Defaults
                to None.
            if_exists (Literal["error", "delete_matching", "overwrite_or_ignore"],
                optional). What to do if the dataset already exists.
        """
        table = pa.Table.from_pandas(df)

        return self.from_arrow(
            table=table,
            schema_name=schema_name,
            table_name=table_name,
            path=path,
            partition_cols=partition_cols,
            if_exists=if_exists,
        )

    def ls(self, path: str) -> Generator[str, None, None]:
        """
        List files and directories under `path`.

        List operation can be slow if there are a lot of objects, hence using a
        generator.

        Args:
            path (str): The path which contents should be listed.

        Yields:
            Generator[str, None, None]: Contents (files and directories) of `path`.
        """
        for obj in self.client.list_objects(self.bucket, path):
            yield obj.object_name

    def rm(self, path: str, recursive: bool = False) -> None:
        """
        Remove a file or directory from MinIO.

        Args:
            path (str): The path to the file to remove.
        """
        if recursive:
            bucket = path.split("/")[2]
            prefix = "/".join(path.split("/")[3:])
            objects = self.client.list_objects(bucket, prefix=prefix, recursive=True)
            for obj in objects:
                self.client.remove_object(bucket, obj.object_name)
            return

        self.client.remove_object(self.bucket, path)

    def _check_if_file_exists(self, path: str) -> bool:
        try:
            self.client.stat_object(self.bucket, path)
            return True
        except S3Error as e:
            if "Object does not exist" in e.message:
                return False
            else:
                raise e

    def check_connection(self) -> None:
        """Verify connectivity to the MinIO endpoint."""
        try:
            self.client.bucket_exists(self.bucket)
        except NewConnectionError as e:
            raise ValueError(
                f"Connection to MinIO endpoint '{self.endpoint}' failed with error: \n{e}",
                "Please check your credentials and try again.",
            )
        except Exception as e:
            raise ValueError(
                f"Connection to MinIO endpoint '{self.endpoint}' failed with error: \n{e}"
            )
        self.logger.info("Connection successful!")



import logging
import re
from collections import OrderedDict
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Tuple,
    Union,
)
from typing import (
    OrderedDict as OrderedDictType,
)

import numpy as np
import pandas as pd

try:
    import pyrfc
    from pyrfc._exception import ABAPApplicationError
except ModuleNotFoundError:
    raise ImportError("pyfrc is required to use the SAPRFC source.")
from sql_metadata import Parser

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError, DataBufferExceeded
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns, validate

logger = logging.getLogger()


def remove_whitespaces(text):
    return " ".join(text.split())


def get_keyword_for_condition(where: str, condition: str) -> str:
    where = where[: where.find(condition)]
    return where.split()[-1]


def get_where_uppercased(where: str) -> str:
    """
    Uppercase a WHERE clause's keywords without
    altering the original string.
    """
    where_and_uppercased = re.sub("\\sand ", " AND ", where)
    where_and_and_or_uppercased = re.sub("\\sor ", " OR ", where_and_uppercased)
    return where_and_and_or_uppercased


def remove_last_condition(where: str) -> str:
    """Remove the last condtion from a WHERE clause."""
    where = get_where_uppercased(where)
    split_by_and = re.split("\\sAND ", where)
    conditions = [re.split("\\sOR ", expr) for expr in split_by_and]
    conditions_flattened = [
        condition for sublist in conditions for condition in sublist
    ]

    condition_to_remove = conditions_flattened[-1]

    where_trimmed = where[: where.find(condition_to_remove)].split()
    where_trimmed_without_last_keyword = " ".join(where_trimmed[:-1])

    return where_trimmed_without_last_keyword, condition_to_remove


def trim_where(where: str) -> Tuple[str, OrderedDictType[str, str]]:
    """
    Trim a WHERE clause to 75 characters or less,
    as required by SAP. The rest of filters will be applied
    in-memory on client side.
    """

    if len(where) <= 75:
        return where, None

    wheres_to_add = OrderedDict()
    keywords_with_conditions = []
    where_trimmed = where
    while len(where_trimmed) > 75:
        # trim the where
        where_trimmed, removed_condition = remove_last_condition(where_trimmed)

        # store the removed conditions so we can readd them later
        keyword = get_keyword_for_condition(where, removed_condition)
        keywords_with_conditions.append((keyword, removed_condition))

    wheres_to_add_sorted = keywords_with_conditions[::-1]
    wheres_to_add = OrderedDict(wheres_to_add_sorted)

    return where_trimmed, wheres_to_add


def detect_extra_rows(
    row_index: int, data_raw: np.array, chunk: int, fields: List[str]
) -> Union[int, np.array, bool]:
    """Check if, in between calls to the SAP table, the number of rows have increased.
        If so, remove the last rows added, to fit the size of the previous columns.

    Args:
        row_index (int): Number of rows set it down in he first iteration with the SAP table.
        data_raw (np.array): Array with the data retrieve from SAP table.
        chunk (int): The part number in which a number of SAP table columns have been split.
        fields (List[str]): A list with the names of the columns in a chunk.

    Returns:
        Union[int, np.array, bool]: A tuple with the parameters "row_index", "data_raw", a new
            boolean variable "start" to indicate when the for loop has to be restarted,
            and "chunk" variable.
    """
    start = False
    if row_index == 0:
        row_index = data_raw.shape[0]
        if row_index == 0:
            logger.warning(
                f"Empty output was generated for chunk {chunk} in columns {fields}."
            )
            start = True
    elif data_raw.shape[0] != row_index:
        data_raw = data_raw[:row_index]
        logger.warning(
            "New rows were generated during the execution of the script. The table is truncated to the number of rows for the first chunk."
        )

    return row_index, data_raw, start


def replace_separator_in_data(
    data_raw: np.array,
    no_sep_index: np.array,
    record_key: str,
    pos_sep_index: np.array,
    sep: str,
    replacement: str,
) -> np.array:
    """Function to replace the extra separator in every row of the data_raw numpy array.

    Args:
        data_raw (np.array): Array with the data retrieve from SAP table.
        no_sep_index (np.array): Array with indexes where are extra separators characters in rows.
        record_key (str): Key word to extract the data from the numpy array "data_raw".
        pos_sep_index (np.array): Array with indexes where are placed real separators.
        sep (str): Which separator to use when querying SAP.
        replacement (str): In case of sep is on a columns, set up a new character to replace
            inside the string to avoid flow breakdowns.

    Returns:
        np.array: the same data_raw numpy array with the "replacement" separator instead.
    """
    for no_sep in no_sep_index:
        logger.warning(
            "A separator character was found and replaced inside a string text that could produce future errors:"
        )
        logger.warning("\n" + data_raw[no_sep][record_key])
        split_array = np.array([*data_raw[no_sep][record_key]])
        position = np.where(split_array == f"{sep}")[0]
        index_sep_index = np.argwhere(np.in1d(position, pos_sep_index) == False)  # noqa
        index_sep_index = index_sep_index.reshape(
            len(index_sep_index),
        )
        split_array[position[index_sep_index]] = replacement
        data_raw[no_sep][record_key] = "".join(split_array)
        logger.warning("\n" + data_raw[no_sep][record_key])

    return data_raw


def catch_extra_separators(
    data_raw: np.array, record_key: str, sep: str, fields: List[str], replacement: str
) -> np.array:
    """Function to replace extra separators in every row of the table.

    Args:
        data_raw (np.array): Array with the data retrieve from SAP table.
        record_key (str): Key word to extract the data from the numpy array "data_raw".
        sep (str): Which separator to use when querying SAP.
        fields (List[str]): A list with the names of the columns in a chunk.
        replacement (str): In case of sep is on a columns, set up a new character to replace
            inside the string to avoid flow breakdowns.

    Returns:
        np.array: The argument "data_raw" with no extra delimiters.
    """

    # remove scape characters from data_raw ("\t")
    for n, r in enumerate(data_raw):
        if "\t" in r[record_key]:
            data_raw[n][record_key] = r[record_key].replace("\t", " ")

    # first it is identified where the data has an extra separator in text columns.
    sep_counts = np.array([], dtype=int)
    for row in data_raw:
        sep_counts = np.append(sep_counts, row[record_key].count(f"{sep}"))

    no_sep_index = np.argwhere(sep_counts != len(fields) - 1)
    no_sep_index = no_sep_index.reshape(
        len(no_sep_index),
    )
    sep_index = np.argwhere(sep_counts == len(fields) - 1)
    sep_index = sep_index.reshape(
        len(sep_index),
    )
    # identifying "good" rows we obtain the index of separator positions.
    pos_sep_index = np.array([], dtype=int)
    for data in data_raw[sep_index]:
        pos_sep_index = np.append(
            pos_sep_index,
            np.where(np.array([*data[record_key]]) == f"{sep}"),
        )
    pos_sep_index = np.unique(pos_sep_index)

    # in rows with an extra separator, we replace them by another character: "-" by default
    data_raw = replace_separator_in_data(
        data_raw,
        no_sep_index,
        record_key,
        pos_sep_index,
        sep,
        replacement,
    )

    return data_raw


def gen_split(data: Iterable[str], sep: str, record_key: str) -> Iterator[List[str]]:
    """
    Splits each string in the given iterable using the specified separator and yields the resulting list.
    Helps to reduce memory usage when processing big data sets.

    Args:
        data: An iterable collection of strings to be split.
        sep: The string separator used to split each string in `data`.
        record_key: A key for extraction of the data from the nested dict.

    Yields:
        A list of substrings for each string in `data`, split using `sep`.

    Examples:
        >>> list(gen_strip(["a|b|c", "d|e|f|g"], "|", "WA"))
        [['a', 'b', 'c'], ['d', 'e', 'f', 'g']]
    """
    for row in data:
        yield row[record_key].split(sep)


class SAPRFC(Source):
    """
    A class for querying SAP with SQL using the RFC protocol.

    Note that only a very limited subset of SQL is supported:
    - aliases
    - where clauses combined using the AND operator
    - limit & offset

    Unsupported:
    - aggregations
    - joins
    - subqueries
    - etc.
    """

    def __init__(
        self,
        sep: str = None,
        func: str = "RFC_READ_TABLE",
        rfc_total_col_width_character_limit: int = 400,
        credentials: Dict[str, Any] = None,
        config_key: str = None,
        *args,
        **kwargs,
    ):
        """Create an instance of the SAPRFC class.

        Args:
            sep (str, optional): Which separator to use when querying SAP. If not provided,
            multiple options are automatically tried.
            func (str, optional): SAP RFC function to use. Defaults to "RFC_READ_TABLE".
            rfc_total_col_width_character_limit (int, optional): Number of characters by which query will be split in chunks
                in case of too many columns for RFC function. According to SAP documentation, the limit is
                512 characters. However, we observed SAP raising an exception even on a slightly lower number
                of characters, so we add a safety margin. Defaults to 400.
            credentials (Dict[str, Any], optional): 'api_key'. Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant credentials.

        Raises:
            CredentialError: If provided credentials are incorrect.
        """

        self._con = None

        credentials = credentials or get_source_credentials(config_key)
        if credentials is None:
            raise CredentialError("Please specify the credentials.")

        super().__init__(*args, credentials=credentials, **kwargs)

        self.sep = sep
        self.client_side_filters = None
        self.func = func
        self.rfc_total_col_width_character_limit = rfc_total_col_width_character_limit

    @property
    def con(self) -> pyrfc.Connection:
        if self._con is not None:
            return self._con
        con = pyrfc.Connection(**self.credentials)
        self._con = con
        return con

    def check_connection(self) -> None:
        self.logger.info("Checking the connection...")
        self.con.ping()
        self.logger.info("Connection has been validated successfully.")

    def close_connection(self) -> None:
        """Closing RFC connection."""
        self.con.close()
        self.logger.info("Connection has been closed successfully.")

    def get_function_parameters(
        self,
        function_name: str,
        description: Union[None, Literal["short", "long"]] = "short",
        *args,
    ) -> Union[List[str], pd.DataFrame]:
        """Get the description for a SAP RFC function.

        Args:
            function_name (str): The name of the function to detail.
            description (Union[None, Literal[, optional): Whether to display
            a short or a long description. Defaults to "short".

        Raises:
            ValueError: If the argument for description is incorrect.

        Returns:
            Union[List[str], pd.DataFrame]: Either a list of the function's
            parameter names (if 'description' is set to None),
            or a short or long description.
        """
        if description is not None:
            if description not in ["short", "long"]:
                raise ValueError(
                    "Incorrect value for 'description'. Correct values: (None, 'short', 'long'"
                )

        descr = self.con.get_function_description(function_name, *args)
        param_names = [param["name"] for param in descr.parameters]
        detailed_params = descr.parameters
        filtered_detailed_params = [
            {
                "name": param["name"],
                "parameter_type": param["parameter_type"],
                "default_value": param["default_value"],
                "optional": param["optional"],
                "parameter_text": param["parameter_text"],
            }
            for param in descr.parameters
        ]

        if description is not None:
            if description == "long":
                params = detailed_params
            else:
                params = filtered_detailed_params
            params = pd.DataFrame.from_records(params)
        else:
            params = param_names

        return params

    def _get_where_condition(self, sql: str) -> str:
        """Retrieve the WHERE conditions from a SQL query.

        Args:
            sql (str): The input SQL query.

        Raises:
            ValueError: Raised if the WHERE clause is longer than
            75 characters (SAP's limitation) and the condition for the
            extra clause(s) is OR.

        Returns:
            str: The where clause trimmed to <= 75 characters.
        """

        where_match = re.search("\\sWHERE ", sql.upper())
        if not where_match:
            return None

        limit_match = re.search("\\sLIMIT ", sql.upper())
        limit_pos = limit_match.span()[0] if limit_match else len(sql)

        where = sql[where_match.span()[1] : limit_pos]
        where_sanitized = remove_whitespaces(where)
        where_trimmed, client_side_filters = trim_where(where_sanitized)
        if client_side_filters:
            self.logger.warning(
                "A WHERE clause longer than 75 character limit detected."
            )
            if "OR" in [key.upper() for key in client_side_filters.keys()]:
                raise ValueError(
                    "WHERE conditions after the 75 character limit can only be combined with the AND keyword."
                )
            else:
                filters_pretty = list(client_side_filters.items())
                self.logger.warning(
                    f"Trimmed conditions ({filters_pretty}) will be applied client-side."
                )
                self.logger.warning("See the documentation for caveats.")

        self.client_side_filters = client_side_filters
        return where_trimmed

    @staticmethod
    def _get_table_name(sql: str) -> str:
        parsed = Parser(sql)
        if len(parsed.tables) > 1:
            raise ValueError("Querying more than one table is not supported.")
        return parsed.tables[0]

    def _build_pandas_filter_query(
        self, client_side_filters: OrderedDictType[str, str]
    ) -> str:
        """Build a WHERE clause that will be applied client-side.
        This is required if the WHERE clause passed to query() is
        longer than 75 characters.

        Args:
            client_side_filters (OrderedDictType[str, str]): The
            client-side filters to apply.

        Returns:
            str: the WHERE clause reformatted to fit the format
            required by DataFrame.query().
        """
        for i, f in enumerate(client_side_filters.items()):
            if i == 0:
                # skip the first keyword; we assume it's "AND"
                query = f[1]
            else:
                query += " " + f[0] + " " + f[1]

            filter_column_name = f[1].split()[0]
            resolved_column_name = self._resolve_col_name(filter_column_name)
        query = re.sub("\\s?=\\s?", " == ", query).replace(
            filter_column_name, resolved_column_name
        )
        return query

    def extract_values(self, sql: str) -> None:
        """TODO: This should cover all values, not just columns"""
        self.where = self._get_where_condition(sql)
        self.select_columns = self._get_columns(sql, aliased=False)
        self.select_columns_aliased = self._get_columns(sql, aliased=True)

    def _resolve_col_name(self, column: str) -> str:
        """Get aliased column name if it exists, otherwise return column name."""
        return self.aliases_keyed_by_columns.get(column, column)

    def _get_columns(self, sql: str, aliased: bool = False) -> List[str]:
        """Retrieve column names from a SQL query.

        Args:
            sql (str): The SQL query to parse.
            aliased (bool, optional): Whether to returned aliased
            names. Defaults to False.

        Returns:
            List[str]: A list of column names.
        """
        parsed = Parser(sql)
        columns = list(parsed.columns_dict["select"])
        if aliased:
            aliases_keyed_by_alias = parsed.columns_aliases
            aliases_keyed_by_columns = OrderedDict(
                {val: key for key, val in aliases_keyed_by_alias.items()}
            )

            self.aliases_keyed_by_columns = aliases_keyed_by_columns

            columns = [
                (
                    aliases_keyed_by_columns[col]
                    if col in aliases_keyed_by_columns
                    else col
                )
                for col in columns
            ]

        if self.client_side_filters:
            # In case the WHERE clause is > 75 characters long, we execute the rest of the filters
            # client-side. To do this, we need to pull all fields in the client-side WHERE conditions.
            # Below code adds these columns to the list of SELECTed fields.
            cols_to_add = [v.split()[0] for v in self.client_side_filters.values()]
            if aliased:
                cols_to_add = [aliases_keyed_by_columns[col] for col in cols_to_add]
            columns.extend(cols_to_add)
            columns = list(dict.fromkeys(columns))  # remove duplicates

        return columns

    @staticmethod
    def _get_limit(sql: str) -> int:
        """Get limit from the query"""
        limit_match = re.search("\\sLIMIT ", sql.upper())
        if not limit_match:
            return None

        return int(sql[limit_match.span()[1] :].split()[0])

    @staticmethod
    def _get_offset(sql: str) -> int:
        """Get offset from the query"""
        offset_match = re.search("\\sOFFSET ", sql.upper())
        if not offset_match:
            return None

        return int(sql[offset_match.span()[1] :].split()[0])

    def query(self, sql: str, sep: str = None) -> None:
        """Parse an SQL query into pyRFC commands and save it into
        an internal dictionary.

        Args:
            sql (str): The SQL query to be ran.
            sep (str, optional): The separator to be used
            to split columns in the result blob. Defaults to self.sep.

        Raises:
            ValueError: If the query is not a SELECT query.
        """

        if not sql.strip().upper().startswith("SELECT"):
            raise ValueError("Only SELECT queries are supported.")

        sep = sep if sep is not None else self.sep

        self.sql = sql

        self.extract_values(sql)

        table_name = self._get_table_name(sql)
        # this has to be called before checking client_side_filters
        where = self.where
        columns = self.select_columns
        character_limit = self.rfc_total_col_width_character_limit
        # due to the RFC_READ_TABLE limit of characters per row, colums are splited into smaller lists
        lists_of_columns = []
        cols = []
        col_length_total = 0
        for col in columns:
            info = self.call("DDIF_FIELDINFO_GET", TABNAME=table_name, FIELDNAME=col)
            col_length = info["DFIES_TAB"][0]["LENG"]
            col_length_total += int(col_length)
            if col_length_total <= character_limit:
                cols.append(col)
            else:
                lists_of_columns.append(cols)
                cols = [col]
                col_length_total = 0
        lists_of_columns.append(cols)

        columns = lists_of_columns
        options = [{"TEXT": where}] if where else None
        limit = self._get_limit(sql)
        offset = self._get_offset(sql)
        query_json = dict(
            QUERY_TABLE=table_name,
            FIELDS=columns,
            OPTIONS=options,
            ROWCOUNT=limit,
            ROWSKIPS=offset,
            DELIMITER=sep,
        )
        # SAP doesn't understand None, so we filter out non-specified parameters
        query_json_filtered = {
            key: query_json[key] for key in query_json if query_json[key] is not None
        }
        self._query = query_json_filtered

    def call(self, func: str, *args, **kwargs):
        """Call a SAP RFC function"""
        return self.con.call(func, *args, **kwargs)

    def _get_alias(self, column: str) -> str:
        return self.aliases_keyed_by_columns.get(column, column)

    def _get_client_side_filter_cols(self):
        return [f[1].split()[0] for f in self.client_side_filters.items()]

    @add_viadot_metadata_columns
    def to_df(self, tests: dict = None):
        """
        Load the results of a query into a pandas DataFrame.

        Due to SAP limitations, if the length of the WHERE clause is longer than 75
        characters, we trim whe WHERE clause and perform the rest of the filtering
        on the resulting DataFrame. Eg. if the WHERE clause contains 4 conditions
        and has 80 characters, we only perform 3 filters in the query, and perform
        the last filter on the DataFrame. If characters per row limit will be exceeded,
        data will be downloaded in chunks.

        Source: https://success.jitterbit.com/display/DOC/Guide+to+Using+RFC_READ_TABLE+to+Query+SAP+Tables#GuidetoUsingRFC_READ_TABLEtoQuerySAPTables-create-the-operation
        - WHERE clause: 75 character limit
        - SELECT: 512 character row limit

        Args:
            tests (Dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.

        Returns:
            pd.DataFrame: A DataFrame representing the result of the query provided in `PyRFC.query()`.
        """
        params = self._query
        columns = self.select_columns_aliased
        sep = self._query.get("DELIMITER")
        fields_lists = self._query.get("FIELDS")
        if len(fields_lists) > 1:
            logger.info(f"Data will be downloaded in {len(fields_lists)} chunks.")
        func = self.func
        if sep is None:
            # automatically find a working separator
            SEPARATORS = [
                "|",
                "/t",
                "#",
                ";",
                "@",
                "%",
                "^",
                "`",
                "~",
                "{",
                "}",
                "$",
            ]
        else:
            SEPARATORS = [sep]

        records = None
        for sep in SEPARATORS:
            logger.info(f"Checking if separator '{sep}' works.")
            df = pd.DataFrame()
            self._query["DELIMITER"] = sep
            chunk = 1
            for fields in fields_lists:
                logger.info(f"Downloading {chunk} data chunk...")
                try:
                    self._query["FIELDS"] = fields
                    try:
                        response = self.call(func, **params)
                    except ABAPApplicationError as e:
                        if e.key == "DATA_BUFFER_EXCEEDED":
                            raise DataBufferExceeded(
                                "Character limit per row exceeded. Please select fewer columns."
                            )
                        else:
                            raise e
                    record_key = "WA"
                    data_raw = response["DATA"]
                    records = [row[record_key].split(sep) for row in data_raw]
                    df[fields] = records
                    chunk += 1
                except ValueError:
                    df = pd.DataFrame()
                    continue
        if not records:
            logger.warning("Empty output was generated.")
            columns = []
        df.columns = columns

        if self.client_side_filters:
            filter_query = self._build_pandas_filter_query(self.client_side_filters)
            df.query(filter_query, inplace=True)
            client_side_filter_cols_aliased = [
                self._get_alias(col) for col in self._get_client_side_filter_cols()
            ]
            cols_to_drop = [
                col
                for col in client_side_filter_cols_aliased
                if col not in self.select_columns_aliased
            ]
            df.drop(cols_to_drop, axis=1, inplace=True)
        self.close_connection()

        if tests:
            validate(df=df, tests=tests)

        return df


class SAPRFCV2(Source):
    """
    A class for querying SAP with SQL using the RFC protocol.

    Note that only a very limited subset of SQL is supported:
    - aliases
    - where clauses combined using the AND operator
    - limit & offset

    Unsupported:
    - aggregations
    - joins
    - subqueries
    - etc.
    """

    def __init__(
        self,
        sep: str = None,
        replacement: str = "-",
        func: str = "RFC_READ_TABLE",
        rfc_total_col_width_character_limit: int = 400,
        rfc_unique_id: List[str] = None,
        credentials: Dict[str, Any] = None,
        config_key: str = None,
        *args,
        **kwargs,
    ):
        """Create an instance of the SAPRFC class.

        Args:
            sep (str, optional): Which separator to use when querying SAP. If not provided,
            multiple options are automatically tried.
            replacement (str, optional): In case of separator is on a columns, set up a new character to replace
                inside the string to avoid flow breakdowns. Defaults to "-".
            func (str, optional): SAP RFC function to use. Defaults to "RFC_READ_TABLE".
            rfc_total_col_width_character_limit (int, optional): Number of characters by which query will be split in chunks
                in case of too many columns for RFC function. According to SAP documentation, the limit is
                512 characters. However, we observed SAP raising an exception even on a slightly lower number
                of characters, so we add a safety margin. Defaults to 400.
            rfc_unique_id  (List[str], optional): Reference columns to merge chunks Data Frames. These columns must to be unique. Defaults to None.
            credentials (Dict[str, Any], optional): 'api_key'. Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant credentials.

        Raises:
            CredentialError: If provided credentials are incorrect.
        """

        self._con = None

        credentials = credentials or get_source_credentials(config_key)
        if credentials is None:
            raise CredentialError("Please specify the credentials.")

        super().__init__(*args, credentials=credentials, **kwargs)

        self.sep = sep
        self.replacement = replacement
        self.client_side_filters = None
        self.func = func
        self.rfc_total_col_width_character_limit = rfc_total_col_width_character_limit

        if rfc_unique_id is not None:
            self.rfc_unique_id = list(set(rfc_unique_id))
        else:
            self.rfc_unique_id = rfc_unique_id

    @property
    def con(self) -> pyrfc.Connection:
        if self._con is not None:
            return self._con
        con = pyrfc.Connection(**self.credentials)
        self._con = con
        return con

    def check_connection(self) -> None:
        self.logger.info("Checking the connection...")
        self.con.ping()
        self.logger.info("Connection has been validated successfully.")

    def close_connection(self) -> None:
        """Closing RFC connection."""
        self.con.close()
        self.logger.info("Connection has been closed successfully.")

    def get_function_parameters(
        self,
        function_name: str,
        description: Union[None, Literal["short", "long"]] = "short",
        *args,
    ) -> Union[List[str], pd.DataFrame]:
        """Get the description for a SAP RFC function.

        Args:
            function_name (str): The name of the function to detail.
            description (Union[None, Literal[, optional): Whether to display
            a short or a long description. Defaults to "short".

        Raises:
            ValueError: If the argument for description is incorrect.

        Returns:
            Union[List[str], pd.DataFrame]: Either a list of the function's
            parameter names (if 'description' is set to None),
            or a short or long description.
        """
        if description is not None:
            if description not in ["short", "long"]:
                raise ValueError(
                    "Incorrect value for 'description'. Correct values: (None, 'short', 'long'"
                )

        descr = self.con.get_function_description(function_name, *args)
        param_names = [param["name"] for param in descr.parameters]
        detailed_params = descr.parameters
        filtered_detailed_params = [
            {
                "name": param["name"],
                "parameter_type": param["parameter_type"],
                "default_value": param["default_value"],
                "optional": param["optional"],
                "parameter_text": param["parameter_text"],
            }
            for param in descr.parameters
        ]

        if description is not None:
            if description == "long":
                params = detailed_params
            else:
                params = filtered_detailed_params
            params = pd.DataFrame.from_records(params)
        else:
            params = param_names

        return params

    def _get_where_condition(self, sql: str) -> str:
        """Retrieve the WHERE conditions from a SQL query.

        Args:
            sql (str): The input SQL query.

        Raises:
            ValueError: Raised if the WHERE clause is longer than
            75 characters (SAP's limitation) and the condition for the
            extra clause(s) is OR.

        Returns:
            str: The where clause trimmed to <= 75 characters.
        """

        where_match = re.search("\\sWHERE ", sql.upper())
        if not where_match:
            return None

        limit_match = re.search("\\sLIMIT ", sql.upper())
        limit_pos = limit_match.span()[0] if limit_match else len(sql)

        where = sql[where_match.span()[1] : limit_pos]
        where_sanitized = remove_whitespaces(where)
        where_trimmed, client_side_filters = trim_where(where_sanitized)
        if client_side_filters:
            self.logger.warning(
                "A WHERE clause longer than 75 character limit detected."
            )
            if "OR" in [key.upper() for key in client_side_filters.keys()]:
                raise ValueError(
                    "WHERE conditions after the 75 character limit can only be combined with the AND keyword."
                )
            else:
                filters_pretty = list(client_side_filters.items())
                self.logger.warning(
                    f"Trimmed conditions ({filters_pretty}) will be applied client-side."
                )
                self.logger.warning("See the documentation for caveats.")

        self.client_side_filters = client_side_filters
        return where_trimmed

    @staticmethod
    def _get_table_name(sql: str) -> str:
        parsed = Parser(sql)
        if len(parsed.tables) > 1:
            raise ValueError("Querying more than one table is not supported.")
        return parsed.tables[0]

    def _build_pandas_filter_query(
        self, client_side_filters: OrderedDictType[str, str]
    ) -> str:
        """Build a WHERE clause that will be applied client-side.
        This is required if the WHERE clause passed to query() is
        longer than 75 characters.

        Args:
            client_side_filters (OrderedDictType[str, str]): The
            client-side filters to apply.

        Returns:
            str: the WHERE clause reformatted to fit the format
            required by DataFrame.query().
        """
        for i, f in enumerate(client_side_filters.items()):
            if i == 0:
                # skip the first keyword; we assume it's "AND"
                query = f[1]
            else:
                query += " " + f[0] + " " + f[1]

            filter_column_name = f[1].split()[0]
            resolved_column_name = self._resolve_col_name(filter_column_name)
        query = re.sub("\\s?=\\s?", " == ", query).replace(
            filter_column_name, resolved_column_name
        )
        return query

    def extract_values(self, sql: str) -> None:
        """TODO: This should cover all values, not just columns"""
        self.where = self._get_where_condition(sql)
        self.select_columns = self._get_columns(sql, aliased=False)
        self.select_columns_aliased = self._get_columns(sql, aliased=True)

    def _resolve_col_name(self, column: str) -> str:
        """Get aliased column name if it exists, otherwise return column name."""
        return self.aliases_keyed_by_columns.get(column, column)

    def _get_columns(self, sql: str, aliased: bool = False) -> List[str]:
        """Retrieve column names from a SQL query.

        Args:
            sql (str): The SQL query to parse.
            aliased (bool, optional): Whether to returned aliased
            names. Defaults to False.

        Returns:
            List[str]: A list of column names.
        """
        parsed = Parser(sql)
        columns = list(parsed.columns_dict["select"])
        if aliased:
            aliases_keyed_by_alias = parsed.columns_aliases
            aliases_keyed_by_columns = OrderedDict(
                {val: key for key, val in aliases_keyed_by_alias.items()}
            )

            self.aliases_keyed_by_columns = aliases_keyed_by_columns

            columns = [
                (
                    aliases_keyed_by_columns[col]
                    if col in aliases_keyed_by_columns
                    else col
                )
                for col in columns
            ]

        if self.client_side_filters:
            # In case the WHERE clause is > 75 characters long, we execute the rest of the filters
            # client-side. To do this, we need to pull all fields in the client-side WHERE conditions.
            # Below code adds these columns to the list of SELECTed fields.
            cols_to_add = [v.split()[0] for v in self.client_side_filters.values()]
            if aliased:
                cols_to_add = [aliases_keyed_by_columns[col] for col in cols_to_add]
            columns.extend(cols_to_add)
            columns = list(dict.fromkeys(columns))  # remove duplicates

        return columns

    @staticmethod
    def _get_limit(sql: str) -> int:
        """Get limit from the query"""
        limit_match = re.search("\\sLIMIT ", sql.upper())
        if not limit_match:
            return None

        return int(sql[limit_match.span()[1] :].split()[0])

    @staticmethod
    def _get_offset(sql: str) -> int:
        """Get offset from the query"""
        offset_match = re.search("\\sOFFSET ", sql.upper())
        if not offset_match:
            return None

        return int(sql[offset_match.span()[1] :].split()[0])

    def query(self, sql: str, sep: str = None) -> None:
        """Parse an SQL query into pyRFC commands and save it into
        an internal dictionary.

        Args:
            sql (str): The SQL query to be ran.
            sep (str, optional): The separator to be used
            to split columns in the result blob. Defaults to self.sep.

        Raises:
            ValueError: If the query is not a SELECT query.
        """

        if not sql.strip().upper().startswith("SELECT"):
            raise ValueError("Only SELECT queries are supported.")

        sep = sep if sep is not None else self.sep

        self.sql = sql

        self.extract_values(sql)

        table_name = self._get_table_name(sql)
        # this has to be called before checking client_side_filters
        where = self.where
        columns = self.select_columns
        lists_of_columns = []
        cols = []
        col_length_total = 0
        if isinstance(self.rfc_unique_id[0], str):
            character_limit = self.rfc_total_col_width_character_limit
            for ref_column in self.rfc_unique_id:
                col_length_reference_column = int(
                    self.call(
                        "DDIF_FIELDINFO_GET",
                        TABNAME=table_name,
                        FIELDNAME=ref_column,
                    )["DFIES_TAB"][0]["LENG"]
                )
                if col_length_reference_column > int(
                    self.rfc_total_col_width_character_limit / 4
                ):
                    raise ValueError(
                        f"{ref_column} can't be used as unique column, too large."
                    )
                local_limit = (
                    self.rfc_total_col_width_character_limit
                    - col_length_reference_column
                )
                if local_limit < character_limit:
                    character_limit = local_limit
        else:
            character_limit = self.rfc_total_col_width_character_limit

        for col in columns:
            info = self.call("DDIF_FIELDINFO_GET", TABNAME=table_name, FIELDNAME=col)
            col_length = info["DFIES_TAB"][0]["LENG"]
            col_length_total += int(col_length)
            if col_length_total <= character_limit:
                cols.append(col)
            else:
                if isinstance(self.rfc_unique_id[0], str) and all(
                    [rfc_col not in cols for rfc_col in self.rfc_unique_id]
                ):
                    for rfc_col in self.rfc_unique_id:
                        if rfc_col not in cols:
                            cols.append(rfc_col)
                lists_of_columns.append(cols)
                cols = [col]
                col_length_total = int(col_length)
        else:
            if isinstance(self.rfc_unique_id[0], str) and all(
                [rfc_col not in cols for rfc_col in self.rfc_unique_id]
            ):
                for rfc_col in self.rfc_unique_id:
                    if rfc_col not in cols:
                        cols.append(rfc_col)
            lists_of_columns.append(cols)

        columns = lists_of_columns
        options = [{"TEXT": where}] if where else None
        limit = self._get_limit(sql)
        offset = self._get_offset(sql)
        query_json = dict(
            QUERY_TABLE=table_name,
            FIELDS=columns,
            OPTIONS=options,
            ROWCOUNT=limit,
            ROWSKIPS=offset,
            DELIMITER=sep,
        )
        # SAP doesn't understand None, so we filter out non-specified parameters
        query_json_filtered = {
            key: query_json[key] for key in query_json if query_json[key] is not None
        }
        self._query = query_json_filtered

    def call(self, func: str, *args, **kwargs):
        """Call a SAP RFC function"""
        return self.con.call(func, *args, **kwargs)

    def _get_alias(self, column: str) -> str:
        return self.aliases_keyed_by_columns.get(column, column)

    def _get_client_side_filter_cols(self):
        return [f[1].split()[0] for f in self.client_side_filters.items()]

    @add_viadot_metadata_columns
    def to_df(self, tests: dict = None):
        """
        Load the results of a query into a pandas DataFrame.

        Due to SAP limitations, if the length of the WHERE clause is longer than 75
        characters, we trim whe WHERE clause and perform the rest of the filtering
        on the resulting DataFrame. Eg. if the WHERE clause contains 4 conditions
        and has 80 characters, we only perform 3 filters in the query, and perform
        the last filter on the DataFrame. If characters per row limit will be exceeded,
        data will be downloaded in chunks.

        Source: https://success.jitterbit.com/display/DOC/Guide+to+Using+RFC_READ_TABLE+to+Query+SAP+Tables#GuidetoUsingRFC_READ_TABLEtoQuerySAPTables-create-the-operation
        - WHERE clause: 75 character limit
        - SELECT: 512 character row limit

        Args:
            tests (Dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.

        Returns:
            pd.DataFrame: A DataFrame representing the result of the query provided in `PyRFC.query()`.
        """

        params = self._query
        columns = self.select_columns_aliased
        sep = self._query.get("DELIMITER")
        fields_lists = self._query.get("FIELDS")
        if len(fields_lists) > 1:
            logger.info(f"Data will be downloaded in {len(fields_lists)} chunks.")
        func = self.func
        if sep is None:
            # automatically find a working separator
            SEPARATORS = [
                "|",
                "/t",
                "#",
                ";",
                "@",
                "%",
                "^",
                "`",
                "~",
                "{",
                "}",
                "$",
            ]
        else:
            SEPARATORS = [sep]

        for sep in SEPARATORS:
            logger.info(f"Checking if separator '{sep}' works.")
            if isinstance(self.rfc_unique_id[0], str):
                # columns only for the first chunk and we add the rest later to avoid name conflicts
                df = pd.DataFrame(columns=fields_lists[0])
            else:
                df = pd.DataFrame()
            self._query["DELIMITER"] = sep
            chunk = 1
            row_index = 0
            for fields in fields_lists:
                logger.info(f"Downloading {chunk} data chunk...")
                self._query["FIELDS"] = fields
                try:
                    response = self.call(func, **params)
                except ABAPApplicationError as e:
                    if e.key == "DATA_BUFFER_EXCEEDED":
                        raise DataBufferExceeded(
                            "Character limit per row exceeded. Please select fewer columns."
                        )
                    else:
                        raise e
                # Check and skip if there is no data returned
                if response["DATA"]:
                    record_key = "WA"
                    data_raw = np.array(response["DATA"])
                    del response

                    # if the reference columns are provided not necessary to remove any extra row.
                    if not isinstance(self.rfc_unique_id[0], str):
                        row_index, data_raw, start = detect_extra_rows(
                            row_index, data_raw, chunk, fields
                        )
                    else:
                        start = False

                    records = [row for row in gen_split(data_raw, sep, record_key)]
                    del data_raw

                    if (
                        isinstance(self.rfc_unique_id[0], str)
                        and not list(df.columns) == fields
                    ):
                        df_tmp = pd.DataFrame(columns=fields)
                        df_tmp[fields] = records
                        # SAP adds whitespaces to the first extracted column value
                        # If whitespace is in unique column it must be removed to make a proper merge
                        for col in self.rfc_unique_id:
                            df_tmp[col] = df_tmp[col].str.strip()
                            df[col] = df[col].str.strip()
                        df = pd.merge(df, df_tmp, on=self.rfc_unique_id, how="outer")
                    else:
                        if not start:
                            df[fields] = records
                        else:
                            df[fields] = np.nan
                    chunk += 1
                elif not response["DATA"]:
                    print("No data returned from SAP.")
        df.columns = columns

        if self.client_side_filters:
            filter_query = self._build_pandas_filter_query(self.client_side_filters)
            df.query(filter_query, inplace=True)
            client_side_filter_cols_aliased = [
                self._get_alias(col) for col in self._get_client_side_filter_cols()
            ]
            cols_to_drop = [
                col
                for col in client_side_filter_cols_aliased
                if col not in self.select_columns_aliased
            ]
            df.drop(cols_to_drop, axis=1, inplace=True)
        self.close_connection()

        if tests:
            validate(df=df, tests=tests)

        return df



from viadot.sources.base import SQL


class SQLite(SQL):
    """A SQLite source

    Args:
        server (str): server string, usually localhost
        db (str): the file path to the db e.g. /home/somedb.sqlite
    """

    def __init__(
        self,
        query_timeout: int = 60,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            driver="/usr/lib/x86_64-linux-gnu/odbc/libsqlite3odbc.so",
            query_timeout=query_timeout,
            **kwargs,
        )
        self.credentials["server"] = "localhost"

    @property
    def conn_str(self):
        """Generate a connection string from params or config.
        Note that the user and password are escapedd with '{}' characters.

        Returns:
            str: The ODBC connection string.
        """
        driver = self.credentials["driver"]
        server = self.credentials["server"]
        db_name = self.credentials["db_name"]

        conn_str = f"DRIVER={{{driver}}};SERVER={server};DATABASE={db_name};"

        return conn_str

    def _check_if_table_exists(self, table: str, schema: str = None) -> bool:
        """Checks if table exists.
        Args:
            table (str): Table name.
            schema (str, optional): Schema name. Defaults to None.
        """
        fqn = f"{schema}.{table}" if schema is not None else table
        exists_query = (
            f"SELECT name FROM sqlite_master WHERE type='table' AND name='{fqn}'"
        )
        exists = bool(self.run(exists_query))
        return exists



import io
from typing import Optional, Union

import pandas as pd
import sharepy
from pydantic import BaseModel, root_validator
from sharepy.errors import AuthError

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.signals import SKIP
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns, cleanup_df, validate


class SharepointCredentials(BaseModel):
    site: str  # Path to sharepoint website (e.g : {tenant_name}.sharepoint.com)
    username: str  # Sharepoint username (e.g username@{tenant_name}.com)
    password: str  # Sharepoint password

    @root_validator(pre=True)
    def is_configured(cls, credentials):
        site = credentials.get("site")
        username = credentials.get("username")
        password = credentials.get("password")

        if not (site and username and password):
            raise CredentialError(
                "'site', 'username', and 'password' credentials are required."
            )
        return credentials


class Sharepoint(Source):
    """
    Download Excel files from Sharepoint.

    Args:
        credentials (SharepointCredentials): Sharepoint credentials.
        config_key (str, optional): The key in the viadot config holding relevant credentials.
    """

    def __init__(
        self,
        credentials: SharepointCredentials = None,
        config_key: Optional[str] = None,
        *args,
        **kwargs,
    ):
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = dict(
            SharepointCredentials(**raw_creds)
        )  # validate the credentials
        super().__init__(*args, credentials=validated_creds, **kwargs)

    def get_connection(self) -> sharepy.session.SharePointSession:
        try:
            connection = sharepy.connect(
                site=self.credentials.get("site"),
                username=self.credentials.get("username"),
                password=self.credentials.get("password"),
            )
        except AuthError:
            site = self.credentials.get("site")
            raise CredentialError(
                f"Could not authenticate to {site} with provided credentials."
            )
        return connection

    def download_file(
        self,
        url: str,
        to_path: str,
    ) -> None:
        """
        Download a file from Sharepoint.

        Args:
            url (str): The URL of the file to be downloaded.
            to_path (str): Where to download the file.

        Example:
            download_file(
                url="https://{tenant_name}.sharepoint.com/sites/{directory}/Shared%20Documents/Dashboard/file",
                to_path="file.xlsx"
            )
        """
        conn = self.get_connection()
        conn.getfile(
            url=url,
            filename=to_path,
        )
        conn.close()

    @add_viadot_metadata_columns
    def to_df(
        self,
        url: str,
        sheet_name: Optional[Union[str, list, int]] = None,
        if_empty: str = "warn",
        tests: dict = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Load an Excel file into a pandas DataFrame.

        Args:
            url (str): The URL of the file to be downloaded.
            sheet_name (Optional[Union[str, list, int]], optional): Strings are used for sheet names.
                Integers are used in zero-indexed sheet positions (chart sheets do not count
                as a sheet position). Lists of strings/integers are used to request multiple sheets.
                Specify None to get all worksheets. Defaults to None.
            if_empty (str, optional): What to do if the file is empty. Defaults to "warn".
            tests (Dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.
            kwargs (dict[str, Any], optional): Keyword arguments to pass to pd.ExcelFile.parse(). Note that
            `nrows` is not supported.

        Returns:
            pd.DataFrame: The resulting data as a pandas DataFrame.
        """

        if "xls" not in url.split(".")[-1]:
            raise ValueError("Only Excel files can be loaded into a DataFrame.")

        if "nrows" in kwargs:
            raise ValueError("Parameter 'nrows' is not supported.")

        conn = self.get_connection()

        self.logger.info(f"Downloading data from {url}...")

        response = conn.get(url)
        bytes_stream = io.BytesIO(response.content)
        excel_file = pd.ExcelFile(bytes_stream)

        if sheet_name:
            df = excel_file.parse(sheet_name, **kwargs)
            df["sheet_name"] = sheet_name
        else:
            sheets: list[pd.DataFrame] = []
            for sheet_name in excel_file.sheet_names:
                sheet = excel_file.parse(sheet_name=sheet_name, **kwargs)
                sheet["sheet_name"] = sheet_name
                sheets.append(sheet)
            df = pd.concat(sheets)

        if df.empty:
            try:
                self._handle_if_empty(if_empty)
            except SKIP:
                return pd.DataFrame()
        else:
            self.logger.info(f"Successfully downloaded {len(df)} of data.")

        df_clean = cleanup_df(df)

        if tests:
            validate(df=df_clean, tests=tests)

        return df_clean



from importlib.util import find_spec

from viadot.sources.azure_data_lake import AzureDataLake
from viadot.sources.cloud_for_customers import CloudForCustomers
from viadot.sources.exchange_rates import ExchangeRates
from viadot.sources.genesys import Genesys
from viadot.sources.minio import MinIO
from viadot.sources.redshift_spectrum import RedshiftSpectrum
from viadot.sources.s3 import S3
from viadot.sources.sharepoint import Sharepoint
from viadot.sources.trino import Trino

__all__ = [
    "AzureDataLake",
    "CloudForCustomers",
    "ExchangeRates",
    "Genesys",
    "MinIO",
    "RedshiftSpectrum",
    "S3",
    "Sharepoint",
    "Trino",
]

if find_spec("pyrfc"):
    from viadot.sources.sap_rfc import SAPRFC, SAPRFCV2  # noqa

    __all__.extend(["SAPRFC", "SAPRFCV2"])

if find_spec("pyspark"):
    from viadot.sources.databricks import Databricks  # noqa

    __all__.append("Databricks")



import re
from copy import deepcopy
from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

import pandas as pd
import requests
from pydantic import BaseModel, SecretStr, root_validator

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.sources.base import Source
from viadot.utils import handle_api_response, validate


class CloudForCustomersCredentials(BaseModel):
    username: str  # eg. username@{tenant_name}.com
    password: SecretStr
    url: Optional[str] = None  # The URL to extract records from.
    report_url: Optional[str] = None  # The URL of a prepared report.

    @root_validator(pre=True)
    def is_configured(cls, credentials):
        username = credentials.get("username")
        password = credentials.get("password")

        if not (username and password):
            raise CredentialError("`username` and `password` credentials are required.")

        return credentials


class CloudForCustomers(Source):
    """Cloud for Customers connector to fetch Odata source.

    Args:
        url (str, optional): The URL to the C4C API. E.g 'https://myNNNNNN.crm.ondemand.com/c4c/v1/'.
        endpoint (str, optional): The API endpoint.
        report_url (str, optional): The URL of a prepared report.
        filter_params (Dict[str, Any], optional): Filtering parameters passed to the request. E.g {"$filter": "AccountID eq '1234'"}.
        More info on:   https://userapps.support.sap.com/sap/support/knowledge/en/2330688
        credentials (CloudForCustomersCredentials, optional): Cloud for Customers credentials.
        config_key (str, optional): The key in the viadot config holding relevant credentials.
    """

    DEFAULT_PARAMS = {"$format": "json"}

    def __init__(
        self,
        url: str = None,
        endpoint: str = None,
        report_url: str = None,
        filter_params: Dict[str, Any] = None,
        credentials: CloudForCustomersCredentials = None,
        config_key: Optional[str] = None,
        *args,
        **kwargs,
    ):
        ## Credentials logic
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = dict(
            CloudForCustomersCredentials(**raw_creds)
        )  # validate the credentials
        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.url = url or self.credentials.get("url")
        self.report_url = report_url or self.credentials.get("report_url")

        self.is_report = bool(self.report_url)
        self.endpoint = endpoint

        if self.url:
            self.full_url = urljoin(self.url, self.endpoint)

        if filter_params:
            filter_params_merged = self.DEFAULT_PARAMS.copy()
            filter_params_merged.update(filter_params)

            self.filter_params = filter_params_merged
        else:
            self.filter_params = self.DEFAULT_PARAMS

    @staticmethod
    def create_metadata_url(url: str) -> str:
        """Creates URL to fetch metadata from.

        Args:
            url (str): The URL to transform to metadata URL.

        Returns:
            meta_url (str): The URL to fetch metadata from.
        """
        start = url.split(".svc")[0]
        url_raw = url.split("?")[0]
        end = url_raw.split("/")[-1]
        meta_url = start + ".svc/$metadata?entityset=" + end
        return meta_url

    def _extract_records_from_report_url(self, report_url: str) -> List[Dict[str, Any]]:
        """Fetches report_url to extract records.

        Args:
            report_url (str): The url to extract records from.

        Returns:
            records (List[Dict[str, Any]]): The records extracted from report_url.
        """
        records = []
        while report_url:
            response = self.get_response(report_url, filter_params=self.filter_params)
            response_json = response.json()
            new_records = self.get_entities(response_json, report_url)
            records.extend(new_records)

            report_url = response_json["d"].get("__next")

        return records

    def _extract_records_from_url(self, url: str) -> List[Dict[str, Any]]:
        """Fetches URL to extract records.

        Args:
            url (str): The URL to extract records from.

        Returns:
            records (List[Dict[str, Any]]): The records extracted from URL.
        """
        tmp_full_url = deepcopy(url)
        tmp_filter_params = deepcopy(self.filter_params)
        records = []
        while url:
            response = self.get_response(tmp_full_url, filter_params=tmp_filter_params)
            response_json = response.json()
            if isinstance(response_json["d"], dict):
                # ODATA v2+ API
                new_records = response_json["d"].get("results")
                url = response_json["d"].get("__next", None)
            else:
                # ODATA v1
                new_records = response_json["d"]
                url = response_json.get("__next", None)

            # prevents concatenation of previous urls with filter_params with the same filter_params
            tmp_filter_params = None
            tmp_full_url = url

            if hasattr(new_records, "__iter__"):
                records.extend(new_records)
        return records

    def extract_records(
        self, url: Optional[str], report_url: Optional[str]
    ) -> List[Dict[str, Any]]:
        """Downloads records from `url` or `report_url` if present.

        Returns:
            records (List[Dict[str, Any]]): The records extracted from URL.
        """
        if self.is_report:
            return self._extract_records_from_report_url(url=report_url)
        else:
            if url:
                full_url = urljoin(url, self.endpoint)
            else:
                full_url = self.full_url
            return self._extract_records_from_url(url=full_url)

    def get_entities(
        self, dirty_json: Dict[str, Any], url: str
    ) -> List[Dict[str, Any]]:
        """Extracts entities from request.json(). Entities represents objects that store information.
           More info on: https://help.sap.com/docs/EAD_HANA/0e60f05842fd41078917822867220c78/0bd1db568fa546d6823d4c19a6b609ab.html

        Args:
            dirty_json (Dict[str, Any]): request.json() dict from response to API.
            url (str): The URL to fetch metadata from.

        Returns:
            entities (List[Dict[str, Any]]): list filled with entities.
        """

        metadata_url = self.create_metadata_url(url)
        column_maper_dict = self.get_property_to_sap_label_dict(metadata_url)
        entities = []
        for element in dirty_json["d"]["results"]:
            new_entity = {}
            for key, object_of_interest in element.items():
                if key not in ["__metadata", "Photo", "", "Picture"]:
                    if "{" not in str(object_of_interest):
                        new_key = column_maper_dict.get(key)
                        if new_key:
                            new_entity[new_key] = object_of_interest
                        else:
                            new_entity[key] = object_of_interest
            entities.append(new_entity)
        return entities

    def get_property_to_sap_label_dict(self, url: str = None) -> Dict[str, str]:
        """Creates Dict that maps Property Name to value of SAP label.
           Property: Properties define the characteristics of the data.
           SAP label: Labels are used for identification and for provision of content information.

        Args:
            url (str, optional): The URL to fetch metadata from.

        Returns:
            Dict[str, str]: Property Name to value of SAP label.
        """

        column_mapping = {}
        if url:
            username = self.credentials.get("username")
            pw = self.credentials.get("password")
            response = requests.get(url, auth=(username, pw))
            for sentence in response.text.split("/>"):
                result = re.search(
                    r'(?<=Name=")([^"]+).+(sap:label=")([^"]+)+', sentence
                )
                if result:
                    key = result.groups(0)[0]
                    val = result.groups(0)[2]
                    column_mapping[key] = val
        return column_mapping

    def get_response(
        self,
        url: str,
        filter_params: Dict[str, Any] = None,
        timeout: tuple = (3.05, 60 * 30),
    ) -> requests.models.Response:
        """Handles requests.

        Args:
            url (str): The url to request to.
            filter_params (Dict[str, Any], optional): Additional parameters like filter, used in case of normal url.
            timeout (tuple, optional): The request time-out. Default is (3.05, 60 * 30).

        Returns:
            requests.models.Response.
        """
        username = self.credentials.get("username")
        pw = self.credentials.get("password")
        response = handle_api_response(
            url=url,
            params=filter_params,
            auth=(username, pw),
            timeout=timeout,
        )
        return response

    def to_df(
        self,
        url: str = None,
        fields: List[str] = None,
        dtype: dict = None,
        tests: dict = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Download a table or report into a pandas DataFrame.

        Args:
            url (str): The URL to extract records from.
            fields (List[str], optional): List of fields to put in DataFrame.
            dtype (dict, optional): The dtypes to use in the DataFrame.
            tests (Dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.
            kwargs: The parameters to pass to DataFrame constructor.

        Returns:
            df (pandas.DataFrame): DataFrame containing the records.
        """
        url = url or self.url
        records = self.extract_records(url=url)
        df = pd.DataFrame(data=records, **kwargs)

        if dtype:
            df = df.astype(dtype)

        if fields:
            return df[fields]

        if tests:
            validate(df=df, tests=tests)

        return df



import json
import os
from typing import Literal, Optional, Union

import pandas as pd

try:
    import pyspark.sql.dataframe as spark
    from delta.tables import *  # noqa
except ModuleNotFoundError:
    raise ImportError("pyspark.sql.dataframe is required to use Databricks source.")

from pydantic import BaseModel, root_validator

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError, TableAlreadyExists, TableDoesNotExist
from viadot.sources.base import Source
from viadot.utils import (
    _cast_df_cols,
    add_viadot_metadata_columns,
    build_merge_query,
    df_snakecase_column_names,
)


class DatabricksCredentials(BaseModel):
    host: str  # The host address of the Databricks cluster.
    port: str = "15001"  # The port on which the cluster is exposed. By default '15001'. For Spark Connect, use the port 443 by default.
    cluster_id: str  # The ID of the Databricks cluster to which to connect.
    org_id: Optional[
        str
    ]  # The ID of the Databricks organization to which the cluster belongs. Not required when using Spark Connect.
    token: str  # The access token which will be used to connect to the cluster.

    @root_validator(pre=True)
    def is_configured(cls, credentials):
        host = credentials.get("host")
        cluster_id = credentials.get("cluster_id")
        token = credentials.get("token")

        if not (host and cluster_id and token):
            raise CredentialError(
                "Databricks credentials are not configured correctly."
            )
        return credentials


class Databricks(Source):
    """
    A class for pulling and manipulating data on Databricks.

    Documentation for Databricks is located at:
    https://docs.microsoft.com/en-us/azure/databricks/

    Parameters
    ----------
    credentials : DatabricksCredentials, optional
        Databricks connection configuration.
    config_key (str, optional): The key in the viadot config holding relevant
        credentials.
    """

    DEFAULT_SCHEMA = "default"

    def __init__(
        self,
        credentials: DatabricksCredentials = None,
        config_key: str = None,
        *args,
        **kwargs,
    ):
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(
            DatabricksCredentials(**raw_creds)
        )  # validate the credentials

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self._session = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._session:
            self._session.stop()
            self._session = None

    @property
    def session(self) -> Union[SparkSession, "DatabricksSession"]:  # noqa
        if self._session is None:
            session = self._create_spark_session()
            self._session = session
            return session
        return self._session

    def _create_spark_session(self):
        """
        Establish a connection to the Databricks cluster.

        Returns:
            SparkSession: A configured SparkSession object.
        """

        db_connect_config = dict(
            host=self.credentials.get("host"),
            token=self.credentials.get("token"),
            cluster_id=self.credentials.get("cluster_id"),
            org_id=self.credentials.get("org_id"),
            port=self.credentials.get("port"),
        )

        with open(os.path.expanduser("~/.databricks-connect"), "w") as f:
            json.dump(db_connect_config, f)

        spark = SparkSession.builder.getOrCreate()  # noqa
        return spark

    def _create_spark_connect_session(self):
        """
        Establish a connection to a Databricks cluster.

        Returns:
            SparkSession: A configured SparkSession object.
        """
        from databricks.connect import DatabricksSession

        workspace_instance_name = self.credentials.get("host").split("/")[2]
        port = self.credentials.get("port")
        token = self.credentials.get("token")
        cluster_id = self.credentials.get("cluster_id")

        conn_str = f"sc://{workspace_instance_name}:{port}/;token={token};x-databricks-cluster-id={cluster_id}"
        spark = DatabricksSession.builder.remote(conn_str).getOrCreate()

        return spark

    @add_viadot_metadata_columns
    def to_df(
        self,
        query: str,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
    ) -> pd.DataFrame:
        """
        Execute a query and return a Pandas DataFrame.

        Args:
            query (str): The query to execute
            if_empty (str, optional): What to do if the query returns no data.
                Defaults to 'warn'.

        Example:
        ```python
        from viadot.sources import Databricks

        databricks = Databricks()
        table_data = databricks.to_df("SELECT * FROM schema.table_1")
        ```
        Returns:
            pd.DataFrame: A Pandas DataFrame containing the requested table's data.
        """
        if query.upper().startswith("SELECT"):
            df = self.run(query, fetch_type="pandas")
            if df.empty:
                self._handle_if_empty(if_empty=if_empty)
        else:
            df = pd.DataFrame()
        return df

    def _pandas_df_to_spark_df(self, df: pd.DataFrame) -> spark.DataFrame:
        """
        Convert a Pandas DataFrame to a Spark DataFrame.

        Args:
            df (pd.DataFrame): The Pandas DataFrame to be converted to a Spark DataFrame.

        Example:
        ```python
        from viadot.sources import Databricks
        import pandas as pd

        databricks = Databricks()
        list = [{"id":"1", "name":"Joe"}]
        df = pd.DataFrame(list)

        spark_df = databricks._pandas_df_to_spark_df(df)
        ```
        Returns:
            spark.DataFrame: The resulting Spark DataFrame.
        """
        spark_df = self.session.createDataFrame(df)
        return spark_df

    def _spark_df_to_pandas_df(self, spark_df: spark.DataFrame) -> pd.DataFrame:
        """
        Convert a Spark DataFrame to a Pandas DataFrame.

        Args:
            df (spark.DataFrame): The Spark DataFrame to be converted to a Pandas DataFrame.

        Example:
        ```python
        from viadot.sources import Databricks

        databricks = Databricks()
        list = [{"id":"1", "name":"Joe"}]
        df = pd.DataFrame(list)
        spark_df = databricks._pandas_df_to_spark_df(df)

        pandas_df = databricks._spark_df_to_pandas_df(spark_df)
        ```
        Returns:
            pd.DataFrame: The resulting Pandas DataFrame.
        """
        return spark_df.toPandas()

    def run(
        self, query: str, fetch_type: Literal["spark", "pandas"] = "spark"
    ) -> Union[spark.DataFrame, pd.DataFrame, bool]:
        """
        Execute an SQL query.

        Args:
            query (str): The query to execute.
            fetch_type (Literal, optional): How to return the data: either
            in the default Spark DataFrame format or as a Pandas DataFrame. Defaults to "spark".

        Example:
        ```python
        from viadot.sources import Databricks

        databricks = Databricks()
        query_result = databricks.run("SELECT * FROM schema.table_1")
        ```
        Returns:
            Union[spark.DataFrame, pd.DataFrame, bool]: Either the result set of a query or,
            in case of DDL/DML queries, a boolean describing whether
            the query was executed successfully.
        """
        if fetch_type not in ["spark", "pandas"]:
            raise ValueError(
                "Only the values 'spark', 'pandas' are allowed for 'fetch_type'"
            )

        query_clean = query.upper().strip()
        query_keywords = ["SELECT", "SHOW", "PRAGMA", "DESCRIBE"]

        query_result = self.session.sql(query)

        if any(query_clean.startswith(word) for word in query_keywords):
            if fetch_type == "spark":
                result = query_result
            else:
                result = self._spark_df_to_pandas_df(query_result)
        else:
            result = True

        return result

    def _check_if_table_exists(self, table: str, schema: str = None) -> bool:
        if schema is None:
            schema = Databricks.DEFAULT_SCHEMA
        return self.session.catalog.tableExists(dbName=schema, tableName=table)

    def _check_if_schema_exists(self, schema: str) -> bool:
        return self.session.catalog.databaseExists(schema)

    def create_table_from_pandas(
        self,
        df: pd.DataFrame,
        table: str,
        schema: str = None,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        if_exists: Literal["replace", "skip", "fail"] = "fail",
        snakecase_column_names: bool = True,
        cast_df_columns: bool = True,
    ) -> bool:
        """
        Create a table using a pandas `DataFrame`.

        Args:
            df (pd.DataFrame): The `DataFrame` to be written as a table.
            table (str): Name of the table to be created.
            schema (str, optional): Name of the schema.
            if_empty (str, optional): What to do if the input `DataFrame` is empty.
                Defaults to 'warn'.
            if_exists (Literal, optional): What to do if the table already exists.
                Defaults to 'fail'.
            snakecase_column_names (bool, optional): Whether to convert column names to snake case.
                Defaults to True.
            cast_df_columns (bool, optional): Converts column types in DataFrame using utils._cast_df_cols().
                This param exists because of possible errors with object cols. Defaults to True.

        Example:
        ```python
        from viadot.sources import Databricks

        databricks = Databricks()
        list = [{"id":"1", "name":"Joe"}]
        df = pd.DataFrame(list)

        new_table = databricks.create_table_from_pandas(
            df=df, schema="viadot_test", table="test"
        )
        ```
        Returns:
            bool: True if the table was created successfully, False otherwise.
        """

        if df.empty:
            self._handle_if_empty(if_empty)

        if schema is None:
            schema = Databricks.DEFAULT_SCHEMA

        if snakecase_column_names:
            df = df_snakecase_column_names(df)

        if cast_df_columns:
            df = _cast_df_cols(df, types_to_convert=["object"])

        fqn = f"{schema}.{table}"
        success_message = f"Table {fqn} has been created successfully."

        if self._check_if_table_exists(schema=schema, table=table):
            if if_exists == "skip":
                self.logger.warning(f"Table {fqn} already exists.")
                result = False
            elif if_exists == "fail":
                raise TableAlreadyExists(fqn)
            elif if_exists == "replace":
                result = self._full_refresh(schema=schema, table=table, df=df)
            else:
                success_message = f"Table {fqn} has been overwritten successfully."
                result = True
        else:
            sdf = self._pandas_df_to_spark_df(df)
            sdf.createOrReplaceTempView("tmp_view")

            result = self.run(
                f"CREATE TABLE {fqn} USING DELTA AS SELECT * FROM tmp_view;"
            )

        if result:
            self.logger.info(success_message)

        return result

    def drop_table(self, table: str, schema: str = None) -> bool:
        """
        Delete an existing table.

        Args:
            schema (str): Name of the schema.
            table (str): Name of the new table to be created.

        Raises:
            TableDoesNotExist: If the table does not exist.

        Example:
        ```python
        from viadot.sources import Databricks

        databricks = Databricks()

        databricks.drop_table(schema="viadot_test", table="test")
        ```
        """
        if schema is None:
            schema = Databricks.DEFAULT_SCHEMA

        fqn = f"{schema}.{table}"

        if self._check_if_table_exists(schema=schema, table=table):
            result = self.run(f"DROP TABLE {fqn}")
            self.logger.info(f"Table {fqn} has been deleted successfully.")
        else:
            raise TableDoesNotExist(fqn=fqn)

        return result

    def _append(self, schema: str, table: str, df: pd.DataFrame):
        fqn = f"{schema}.{table}"
        spark_df = self._pandas_df_to_spark_df(df)
        spark_df.write.format("delta").mode("append").saveAsTable(fqn)

        self.logger.info(f"Table {fqn} has been appended successfully.")

    def _full_refresh(self, schema: str, table: str, df: pd.DataFrame) -> bool:
        """
        Overwrite an existing table with data from a Pandas DataFrame.

        Args:
            schema (str): Name of the schema.
            table (str): Name of the new table to be created.
            df (pd.DataFrame): DataFrame to be used to overwrite the table.

        Example:
        ```python
        from viadot.sources import Databricks

        databricks = Databricks()
        list = [{"id":"1", "name":"Joe"}]
        df = pd.DataFrame(list)

        databricks.insert_into( df=df, schema="viadot_test", table="test", mode="replace")
        ```
        Returns:
            bool: True if the table has been refreshed successfully, False otherwise.
        """
        fqn = f"{schema}.{table}"
        data = self._pandas_df_to_spark_df(df)
        data.write.format("delta").mode("overwrite").option(
            "overwriteSchema", "true"
        ).saveAsTable(fqn)

        self.logger.info(f"Table {fqn} has been refreshed successfully.")

        return True

    def _upsert(
        self,
        df: pd.DataFrame,
        table: str,
        primary_key: str,
        schema: str = None,
    ):
        spark_df = self._pandas_df_to_spark_df(df)
        merge_query = build_merge_query(
            df=spark_df,
            schema=schema,
            table=table,
            primary_key=primary_key,
            source=self,
        )
        result = self.run(merge_query)

        self.logger.info("Data has been upserted successfully.")

        return result

    def insert_into(
        self,
        df: pd.DataFrame,
        table: str,
        schema: str = None,
        primary_key: str = None,
        mode: Literal["replace", "append", "update"] = "append",
    ) -> None:
        """
        Insert data from a pandas `DataFrame` into a Delta table.

        Args:
            df (pd.DataFrame): DataFrame with the data to be inserted into the table.
            table (str): Name of the new table to be created.
            schema (str, Optional): Name of the schema.
            primary_key (str, Optional): The primary key on which the data will be joined.
                Required only when updating existing data.
            mode (str, Optional): Which operation to run with the data. Allowed operations
                are: 'replace', 'append', and 'update'. By default, 'append'.

        Example:
        ```python
        from viadot.sources import Databricks

        databricks = Databricks()
        list = [{"id":"1", "name":"Joe"}]
        df = pd.DataFrame(list)

        databricks.insert_into(
            df=df, schema="viadot_test", table="test", primary_key="pk", mode="update"
        )
        ```
        """
        if schema is None:
            schema = Databricks.DEFAULT_SCHEMA

        fqn = f"{schema}.{table}" if schema else table

        exists = self._check_if_table_exists(schema=schema, table=table)
        if exists:
            if mode == "replace":
                self._full_refresh(df=df, schema=schema, table=table)
            elif mode == "append":
                self._append(df=df, schema=schema, table=table)
            elif mode == "update":
                self._upsert(df=df, schema=schema, table=table, primary_key=primary_key)
            else:
                raise ValueError(
                    "`mode` must be one of: 'replace', 'append', or 'update'."
                )
        else:
            raise ValueError(f"Table {fqn} does not exist.")

    def create_schema(self, schema_name: str) -> bool:
        """
        Create a schema for storing tables.

        Args:
            schema_name (str): Name of the new schema to be created.

        Example:
        ```python
        from viadot.sources import Databricks

        databricks = Databricks()

        databricks.create_schema("schema_1")
        ```
        """
        result = self.run(f"CREATE SCHEMA IF NOT EXISTS {schema_name}")
        self.logger.info(f"Schema {schema_name} has been created successfully.")
        return result

    def drop_schema(self, schema_name: str) -> bool:
        """
        Delete a schema.

        Args:
            schema_name (str): Name of the schema to be deleted.

        Example:
        ```python
        from viadot.sources import Databricks

        databricks = Databricks()

        databricks.drop_schema("schema_1")
        ```
        """
        result = self.run(f"DROP SCHEMA {schema_name}")
        self.logger.info(f"Schema {schema_name} deleted.")
        return result

    def discover_schema(self, table: str, schema: str = None) -> dict:
        """
        Return a table's schema.

        Args:
            schema (str): Name of the schema.
            table (str): Name of the new table to be created.

        Example:
        ```python
        from viadot.sources import Databricks

        databricks = Databricks()

        databricks.discover_schema(schema="viadot_test", table="test")
        ```
        Returns:
            schema (dict): A dictionary containing the schema details of the table.
        """
        if schema is None:
            schema = Databricks.DEFAULT_SCHEMA

        fqn = f"{schema}.{table}"
        result = self.run(f"DESCRIBE {fqn}", fetch_type="pandas")

        # Delete the last 3 records, which are boilerplate for Delta Tables
        col_names = result["col_name"].values.tolist()
        col_names = col_names[:-3]

        data_types = result["data_type"].values.tolist()
        data_types = data_types[:-3]

        schema = dict(zip(col_names, data_types))

        return schema

    def get_table_version(self, table: str, schema: str = None) -> int:
        """
        Get the provided table's version number.

        Args:
            schema (str): Name of the schema.
            table (str): Name of the table to rollback.

        Returns:
            version_number (int): The table's version number.
        ```
        """
        if schema is None:
            schema = Databricks.DEFAULT_SCHEMA

        fqn = f"{schema}.{table}"

        # Get the info regarding the current version of the delta table
        history = self.run(f"DESCRIBE HISTORY {fqn}", "pandas")

        # Extract the current version number
        version_number = history["version"].iat[0]
        return int(version_number)

    def rollback(self, table: str, version_number: int, schema: str = None) -> bool:
        """
        Rollback a table to a previous version.

        Args:
            schema (str): Name of the schema.
            table (str): Name of the table to rollback.
            version_number (int): Number of the table's version to rollback to.

        Example:
        ```python
        from viadot.sources import Databricks
        databricks = Databricks()

        schema = "viadot_test"
        table = "table_1"

        version_number = databricks.get_table_version(schema=schema, table=table)

        # Perform changes on the table, in this example we are appending to the table
        list = [{"id":"1", "name":"Joe"}]
        df = pd.DataFrame(list)
        databricks.insert_into(df=df, schema=schema, table=table)

        databricks.rollback(schema=schema, table=table, version_number=version_number)
        ```
        Returns:
            result (bool): A boolean indicating the success of the rollback.
        """

        if schema is None:
            schema = Databricks.DEFAULT_SCHEMA

        fqn = f"{schema}.{table}"

        # Retrieve the data from the previous table
        old_table = self.to_df(f"SELECT * FROM {fqn}@v{version_number}")

        # Perform full-refresh and overwrite the table with the new data
        result = self.insert_into(
            df=old_table, schema=schema, table=table, mode="replace"
        )

        self.logger.info(
            f"Rollback for table {fqn} to version #{version_number} completed."
        )

        return result



import os
from typing import Any, Dict, List

import pandas as pd
from adlfs import AzureBlobFileSystem, AzureDatalakeFileSystem

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns


class AzureDataLake(Source):
    """
    A class for pulling data from the Azure Data Lakes (gen1 and gen2).

    You can either connect to the lake in general
    (`lake = AzureDataLake(); lake.exists("a/b/c.csv")`),
    or to a particular path (`lake = AzureDataLake(path="a/b/c.csv"); lake.exists()`)

    Args:
        credentials (Dict[str, Any], optional): A dictionary containing
            the following credentials: `account_name`, `tenant_id`,
            `client_id`, and `client_secret`.
        config_key (str, optional): The key in the viadot config holding
            relevant credentials.
    """

    def __init__(
        self,
        path: str = None,
        gen: int = 2,
        credentials: Dict[str, Any] = None,
        config_key: str = None,
        *args,
        **kwargs,
    ):
        credentials = credentials or get_source_credentials(config_key)
        required_credentials = (
            "account_name",
            "tenant_id",
            "client_id",
            "client_secret",
        )
        required_credentials_are_provided = all(
            [rc in credentials for rc in required_credentials]
        )

        if credentials is None:
            raise CredentialError("Please provide the credentials.")
        elif not required_credentials_are_provided:
            raise CredentialError("Please provide all required credentials.")

        super().__init__(*args, credentials=credentials, **kwargs)

        storage_account_name = self.credentials["account_name"]
        tenant_id = self.credentials["tenant_id"]
        client_id = self.credentials["client_id"]
        client_secret = self.credentials["client_secret"]

        self.path = path
        self.gen = gen
        self.storage_options = {
            "tenant_id": tenant_id,
            "client_id": client_id,
            "client_secret": client_secret,
        }
        if gen == 1:
            self.fs = AzureDatalakeFileSystem(
                store_name=storage_account_name,
                tenant_id=tenant_id,
                client_id=client_id,
                client_secret=client_secret,
            )
            self.base_url = f"adl://{storage_account_name}"
        elif gen == 2:
            self.storage_options["account_name"] = storage_account_name
            self.fs = AzureBlobFileSystem(
                account_name=storage_account_name,
                tenant_id=tenant_id,
                client_id=client_id,
                client_secret=client_secret,
            )
            self.base_url = "az://"

    def upload(
        self,
        from_path: str,
        to_path: str = None,
        recursive: bool = False,
        overwrite: bool = False,
    ) -> None:
        """
        Upload file(s) to the lake.

        Args:
            from_path (str): Path to the local file(s) to be uploaded.
            to_path (str): Path to the destination file/folder
            recursive (bool): Set this to true if working with directories.
            overwrite (bool): Whether to overwrite the file(s) if they exist.

        Example:
        ```python
        from viadot.sources import AzureDataLake

        lake = AzureDataLake()
        lake.upload(from_path='tests/test.csv', to_path="sandbox/test.csv")
        ```
        """

        if self.gen == 1:
            raise NotImplementedError(
                "Azure Data Lake Gen1 does not support simple file upload."
            )

        to_path = to_path or self.path

        self.logger.info(f"Uploading file(s) from '{from_path}' to '{to_path}'...")

        try:
            self.fs.upload(
                lpath=from_path,
                rpath=to_path,
                recursive=recursive,
                overwrite=overwrite,
            )
        except FileExistsError:
            # Show a useful error message.
            if recursive:
                msg = f"At least one file in '{to_path}' already exists. Specify `overwrite=True` to overwrite."  # noqa
            else:
                msg = f"The file '{to_path}' already exists. Specify `overwrite=True` to overwrite."
            raise FileExistsError(msg)

        self.logger.info(
            f"Successfully uploaded file(s) from '{from_path}' to '{to_path}'."
        )

    def exists(self, path: str = None) -> bool:
        """
        Check if a location exists in Azure Data Lake.

        Args:
            path (str): The path to check. Can be a file or a directory.

        Example:
        ```python
        from viadot.sources import AzureDataLake

        lake = AzureDataLake(gen=1)
        lake.exists("tests/test.csv")
        ```

        Returns:
            bool: Whether the paths exists.
        """
        path = path or self.path
        return self.fs.exists(path)

    def download(
        self,
        to_path: str,
        from_path: str = None,
        recursive: bool = False,
        overwrite: bool = True,
    ) -> None:
        if overwrite is False:
            raise NotImplementedError(
                "Currently, only the default behavior (overwrite) is available."
            )

        from_path = from_path or self.path
        self.fs.download(rpath=from_path, lpath=to_path, recursive=recursive)

    @add_viadot_metadata_columns
    def to_df(
        self,
        path: str = None,
        sep: str = "\t",
        quoting: int = 0,
        lineterminator: str = None,
        error_bad_lines: bool = None,
    ):
        if quoting is None:
            quoting = 0

        path = path or self.path
        url = os.path.join(self.base_url, path)

        if url.endswith(".csv"):
            df = pd.read_csv(
                url,
                storage_options=self.storage_options,
                sep=sep,
                quoting=quoting,
                lineterminator=lineterminator,
                error_bad_lines=error_bad_lines,
            )
        elif url.endswith(".parquet"):
            df = pd.read_parquet(url, storage_options=self.storage_options)
        else:
            raise ValueError("Only CSV and parquet formats are supported.")

        return df

    def ls(self, path: str = None) -> List[str]:
        """
        Returns a list of files in a path.

        Args:
            path (str, optional): Path to a folder. Defaults to None.
        """
        path = path or self.path
        return self.fs.ls(path)

    def rm(self, path: str = None, recursive: bool = False):
        """
        Deletes files in a path.

        Args:
            path (str, optional): Path to a folder. Defaults to None.
            recursive (bool, optional): Whether to delete files recursively.
                Defaults to False.
        """
        path = path or self.path
        self.fs.rm(path, recursive=recursive)

    def cp(self, from_path: str = None, to_path: str = None, recursive: bool = False):
        """
        Copies the contents of `from_path` to `to_path`.

        Args:
            from_path (str, optional): Path from which to copy file(s).
                Defauls to None.
            to_path (str, optional): Path where to copy file(s). Defaults
                to None.
            recursive (bool, optional): Whether to copy file(s) recursively.
                Defaults to False.
        """
        from_path = from_path or self.path
        to_path = to_path
        self.fs.cp(from_path, to_path, recursive=recursive)

    def from_df(
        self, df: pd.DataFrame, path: str = None, overwrite: bool = False
    ) -> None:
        """
        Upload a pandas `DataFrame` to a file on Azure Data Lake.

        Args:
            df (pd.DataFrame): The pandas `DataFrame` to upload.
            path (str, optional): The destination path. Defaults to None.
            overwrite (bool): Whether to overwrite the file if it exist.
        """

        path = path or self.path

        extension = path.split(".")[-1]
        if extension not in ("csv", "parquet"):
            if "." not in path:
                msg = "Please provide the full path to the file."
            else:
                msg = "Accepted file formats are 'csv' and 'parquet'."
            raise ValueError(msg)

        file_name = path.split("/")[-1]
        if extension == "csv":
            # Can do it simply like this if ADLS accesses are set up correctly
            # url = os.path.join(self.base_url, path)
            # df.to_csv(url, storage_options=self.storage_options)
            df.to_csv(file_name, index=False)
        else:
            df.to_parquet(file_name, index=False)

        self.upload(from_path=file_name, to_path=path, overwrite=overwrite)

        os.remove(file_name)



from __future__ import annotations

import json
import logging
from os.path import expanduser, join
from typing import Optional

from yaml import safe_load

logger = logging.getLogger(__name__)

USER_HOME = expanduser("~")


class Config(dict):
    @classmethod
    def _get_configuration(
        cls, config: dict, key: Optional[str] = None
    ) -> Optional[dict]:
        # Empty config.
        if not config:
            return

        # Config key does not exist or has no values.
        if key:
            config = config.get(key)
            if not config:
                logger.warning(f"No configuration found under the '{key}' config key.")
                return

        return cls(**config)

    @classmethod
    def from_json(cls, path: str, key: Optional[str] = None) -> Config:
        with open(path) as f:
            config = json.load(f)
            return cls._get_configuration(config, key=key)

    @classmethod
    def from_yaml(cls, path: str, key: Optional[str] = None) -> Config:
        with open(path) as f:
            config = safe_load(stream=f)
            return cls._get_configuration(config, key=key)


config_dir = join(USER_HOME, ".config", "viadot")

try:
    CONFIG = Config.from_yaml(join(config_dir, "config.yaml"))
except FileNotFoundError:
    try:
        CONFIG = Config.from_json(join(config_dir, "config.json"))
    except FileNotFoundError:
        CONFIG = Config()
except ValueError:
    # Incorrect config structure.
    logger.warning(f"Could not read default viadot config file from '{config_dir}'.")
    CONFIG = Config()


def get_source_config(key, config=CONFIG):
    source_configs = config.get("sources")
    if source_configs is not None:
        for source_config in source_configs:
            if key in source_config.keys():
                return source_configs[source_configs.index(source_config)][key]


def get_source_credentials(key, config=CONFIG):
    config = get_source_config(key, config)
    if config is not None:
        return config.get("credentials")



import os
from typing import List, Literal, Optional, Tuple

import awswrangler as wr
import boto3
import pandas as pd
import redshift_connector
from pydantic import BaseModel, root_validator

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.sources.base import Source


class RedshiftSpectrumCredentials(BaseModel):
    region_name: str  # The name of the AWS region.
    aws_access_key_id: str  # The AWS access key ID.
    aws_secret_access_key: str  # The AWS secret access key.
    profile_name: str = None  # The name of the IAM profile to use.

    # Below credentials are required only by some methods.
    #
    # The name of a AWS Secret holding credentials to a Redshift instance.
    # This AWS Secret has to include the following configuration:
    # host: Optional[str]
    # port: Optional[str] = "5439"
    # username: Optional[str]
    # password: Optional[str]
    # engine: Optional[str] = "redshift"
    # dbname: Optional[str]
    credentials_secret: Optional[str]
    iam_role: Optional[str]  # The IAM role to assume. Used by `create_schema()`.

    @root_validator(pre=True)
    def is_configured(cls, credentials):
        profile_name = credentials.get("profile_name")
        region_name = credentials.get("region_name")
        aws_access_key_id = credentials.get("aws_access_key_id")
        aws_secret_access_key = credentials.get("aws_secret_access_key")

        profile_credential = profile_name and region_name
        direct_credential = aws_access_key_id and aws_secret_access_key and region_name

        if not (profile_credential or direct_credential):
            raise CredentialError(
                "Either `profile_name` and `region_name`, or `aws_access_key_id`, "
                "`aws_secret_access_key`, and `region_name` must be specified."
            )
        return credentials


class RedshiftSpectrum(Source):
    """
    A class for pulling data from and uploading to a specified Amazon Redshift Spectrum
    external schema.

    Note that internally, AWS SDK refers to schemas as "databases", as external schemas
    correspond to AWS Glue databases. However, to keep consistent naming with all other
    viadot sources, we use the word "schema" instead.

    Args:
        credentials (RedshiftSpectrumCredentials, optional): RedshiftSpectrumCredentials
            credentials. Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.

    Examples:
        ```python
        from viadot.sources import RedshiftSpectrum

        with RedshiftSpectrum(config_key="redshift_spectrum") as redshift:
            redshift.get_schemas()
        ```
    """

    def __init__(
        self,
        credentials: RedshiftSpectrumCredentials = None,
        config_key: str = None,
        *args,
        **kwargs,
    ):
        raw_creds = (
            credentials
            or get_source_credentials(config_key)
            or self._get_env_credentials()
        )
        validated_creds = dict(
            RedshiftSpectrumCredentials(**raw_creds)
        )  # validate the credentials

        super().__init__(*args, credentials=validated_creds, **kwargs)

        if not self.credentials:
            self.logger.debug(
                "Credentials not specified. Falling back to `boto3` default credentials."
            )

        self._session = None
        self._con = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._con:
            self._con.close()
            self._con = None

    @property
    def session(self) -> boto3.session.Session:
        """
        A singleton-like property for initiating an AWS session with boto3.

        Note that this is not an actual session, so it does not need to be closed.
        """
        if not self._session:
            self._session = boto3.session.Session(
                region_name=self.credentials.get("region_name"),
                profile_name=self.credentials.get("profile_name"),
                aws_access_key_id=self.credentials.get("aws_access_key_id"),
                aws_secret_access_key=self.credentials.get("aws_secret_access_key"),
            )
        return self._session

    @property
    def con(self) -> redshift_connector.Connection:
        if not self._con:
            if self.credentials.get("credentials_secret"):
                self._con = wr.redshift.connect(
                    boto3_session=self.session,
                    timeout=10,
                    secret_id=self.credentials.get("credentials_secret"),
                )
            else:
                raise ValueError(
                    "`credentials_secret` config is required to connect to Redshift."
                )
        return self._con

    def _get_env_credentials(self):
        credentials = {
            "region_name": os.environ.get("AWS_DEFAULT_REGION"),
            "aws_access_key_id": os.environ.get("AWS_ACCESS_KEY_ID"),
            "aws_secret_access_key": os.environ.get("AWS_SECRET_ACCESS_KEY"),
        }
        return credentials

    def from_df(
        self,
        df: pd.DataFrame,
        to_path: str,
        schema: str,
        table: str,
        extension: Literal[".parquet", ".csv"] = ".parquet",
        if_exists: Literal["overwrite", "overwrite_partitions", "append"] = "overwrite",
        partition_cols: List[str] = None,
        sep: str = ",",
        description: str = None,
        **kwargs,
    ) -> None:
        """
        Upload a pandas `DataFrame` into a CSV or Parquet file in a specified external
        Amazon Redshift Spectrum schema.

        For a full list of available parameters, please refer to the official documentation:
        https://aws-sdk-pandas.readthedocs.io/en/3.0.0/stubs/awswrangler.s3.to_parquet.html
        https://aws-sdk-pandas.readthedocs.io/en/3.0.0/stubs/awswrangler.s3.to_csv.html

        Args:
            df (pd.DataFrame): Pandas `DataFrame`.
            to_path (str): Path to Amazon S3 folder where the table will be located. If needed,
                a bottom-level directory named f"{table}" is automatically created, so
                that files are always located in a folder named the same as the table.
            schema (str): The name of the schema.
            table (str): The name of the table to load the data into.
            extension (Literal[".parquet", ".csv"], optional): Required file type.
                Defaults to '.parquet'.
            if_exists (Literal["overwrite", "overwrite_partitions", "append"], optional):
                'overwrite' to recreate the table, 'overwrite_partitions' to only recreate
                the partitions, 'append' to append the data. Defaults to 'overwrite'.
            partition_cols (List[str], optional): List of column names that will be used to
                create partitions. Only takes effect if dataset=True. Defaults to None.
            sep (str, optional): Field delimiter for the output file. Defaults to ','.
            description (str, optional): Amazon Redshift Spectrum table description.
                Defaults to None.
        """

        # Ensure files are in a directory named {table}.
        if not to_path.rstrip("/").endswith(table):
            to_path = os.path.join(to_path, table)

        if extension == ".parquet":
            wr.s3.to_parquet(
                boto3_session=self.session,
                df=df,
                path=to_path,
                mode=if_exists,
                dataset=True,
                database=schema,
                table=table,
                partition_cols=partition_cols,
                description=description,
                **kwargs,
            )
        elif extension == ".csv":
            wr.s3.to_csv(
                boto3_session=self.session,
                df=df,
                path=to_path,
                dataset=True,
                database=schema,
                table=table,
                sep=sep,
                **kwargs,
            )
        else:
            raise ValueError("Only CSV and parquet formats are supported.")

    def to_df(
        self,
        schema: str,
        table: str,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Reads a table from an Amazon Redshift Spectrum external schema into a pandas `DataFrame`.
        For a full list of available parameters, please refer to the official documentation:
        https://aws-sdk-pandas.readthedocs.io/en/3.0.0/stubs/awswrangler.s3.read_parquet_table.html

        Args:
            schema (str): The name of the schema.
            table (str): The name of the table to load.
        """

        df = wr.s3.read_parquet_table(
            boto3_session=self.session,
            database=schema,
            table=table,
            **kwargs,
        )

        return df

    def drop_table(
        self,
        schema: str,
        table: str,
        remove_files: bool = True,
    ) -> None:
        """
        Drops a table from a specified Amazon Redshift Spectrum external schema,
        including related files from Amazon S3, if specified.

        Args:
            schema (str): The name of the schema.
            table (str): The name of the table to drop.
            remove_files (bool, optional): If True, Amazon S3 file related to the table
                will be removed. Defaults to True.
        """
        if remove_files:
            table_location = wr.catalog.get_table_location(
                boto3_session=self.session,
                database=schema,
                table=table,
            )
            wr.s3.delete_objects(boto3_session=self.session, path=table_location)

        wr.catalog.delete_table_if_exists(
            boto3_session=self.session, database=schema, table=table
        )

    def get_tables(
        self,
        schema: str,
    ) -> List[str]:
        """
        Returns a list of tables in a specified schema.

        Args:
            schema (str): The name of the schema.
        """
        get_tables_query = f"SELECT t.tablename FROM SVV_EXTERNAL_TABLES t WHERE t.schemaname = '{schema}'"
        with self.con.cursor() as cursor:
            tables_info = cursor.execute(get_tables_query).fetchall()
        return [table_info[0] for table_info in tables_info]

    def _check_if_table_exists(self, schema: str, table: str) -> bool:
        """
        Check if a table exists in a specified Amazon Redshift Spectrum external schema.

        Args:
            schema (str): The name of the schema.
            table (str): The name of the table to verify.

        Returns:
            bool: Whether the table exists.
        """
        return table in self.get_tables(schema=schema)

    def create_schema(
        self,
        schema: str,
        description: Optional[str] = None,
    ) -> None:
        """
        Create an external schema in Amazon Redshift Spectrum.

        This involves two steps:
        - creating a Glue database
        - creating an external schema in Redshift, pointing to above Glue database

        Args:
            schema (str): The name of the schema.
            description (str, optional): The description of the schema. Defaults to None.
        """
        self._create_glue_database(
            database=schema, description=description, exist_ok=True
        )
        create_schema_query = f"""
create external schema if not exists "{schema}" from data catalog
database '{schema}'
iam_role '{self.credentials.get("iam_role")}'
region '{self.credentials.get("region_name")}'
CREATE EXTERNAL DATABASE IF NOT EXISTS;
"""
        with self.con.cursor() as cursor:
            cursor.execute(create_schema_query)
            self.con.commit()

    def _create_glue_database(
        self,
        database: str,
        description: Optional[str] = None,
        exist_ok: bool = False,
    ):
        """Create an AWS Glue database.

        Args:
            database (str): The name of the database.
            description (Optional[str], optional): The description of the database.
                Defaults to None.
            exist_ok (bool, optional): Whether to skip if the database already exists.
                If set to False, will throw `AlreadyExistsException`. Defaults to False.
        """
        wr.catalog.create_database(
            name=database,
            description=description,
            boto3_session=self.session,
            exist_ok=exist_ok,
        )

    def get_schemas(self) -> list[str]:
        """Returns a list of schemas in the current Redshift Spectrum database."""

        # External Redshift schemas
        get_schemas_query = "SELECT schemaname FROM SVV_EXTERNAL_SCHEMAS"
        with self.con.cursor() as cursor:
            schema_names: Tuple[list] = cursor.execute(get_schemas_query).fetchall()
        external_schemas = [schema_name[0] for schema_name in schema_names]

        # Glue databases.
        schema_infos = wr.catalog.get_databases(boto3_session=self.session)
        glue_schemas = [schema_info["Name"] for schema_info in schema_infos]

        # An external Redshift schema is a Spectrum schema only if it's also a Glue database.
        return [schema for schema in external_schemas if schema in glue_schemas]

    def _check_if_schema_exists(self, schema: str) -> bool:
        """
        Check if a schema exists in Amazon Redshift Spectrum.

        Args:
            schema (str): The name of the schema.

        Returns:
            bool: Whether the schema exists.
        """
        return schema in self.get_schemas()

    def _is_spectrum_schema(self, schema: str) -> bool:
        """
        Check if a Redshift schema is a Spectrum schema.

        Args:
            schema (str): The name of the schema.

        Returns:
            bool: Whether the schema is a Spectrum schema.
        """
        return self._check_if_schema_exists(schema)

    def drop_schema(self, schema: str, drop_glue_database: bool = False) -> None:
        """
        Drop a Spectrum schema. If specified, also drop the underlying Glue database.

        Args:
            schema (str): The name of the schema.
        """

        if not self._is_spectrum_schema(schema):
            raise ValueError(f"Schema {schema} is not a Spectrum schema.")

        drop_external_schema_query = f"DROP SCHEMA IF EXISTS {schema}"
        with self.con.cursor() as cursor:
            cursor.execute(drop_external_schema_query)
            self.con.commit()

        if drop_glue_database:
            wr.catalog.delete_database(
                name=schema,
                boto3_session=self.session,
            )



import re
import warnings
from contextlib import contextmanager
from typing import Generator, Literal, Optional

import pandas as pd
import pyarrow as pa
from pydantic import BaseModel, Field
from sqlalchemy import create_engine, text
from sqlalchemy.engine import Connection
from sqlalchemy.exc import SADeprecationWarning
from trino.auth import BasicAuthentication

from viadot.config import get_source_credentials
from viadot.sources.base import Source
from viadot.utils import get_fqn

# Triggered by trino lib
warnings.filterwarnings("ignore", category=SADeprecationWarning)


class TrinoCredentials(BaseModel):
    http_scheme: str = "https"
    host: str = "localhost"
    port: int = 443
    user: str
    password: Optional[str] = None
    catalog: str
    schema_name: Optional[str] = Field(None, alias="schema")
    verify: bool = True


class Trino(Source):
    """
    A class for interacting with Trino as a database. Currently supports only generic
    and Iceberg operations.

    Args:
        credentials (TrinoCredentials): Trino credentials.
        config_key (str, optional): The key in the viadot config holding relevant credentials.
    """

    def __init__(
        self,
        credentials: TrinoCredentials = None,
        config_key: str = None,
        *args,
        **kwargs,
    ):
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = TrinoCredentials(**raw_creds).dict(
            by_alias=True
        )  # validate the credentials

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.http_scheme = self.credentials.get("http_scheme")
        self.host = self.credentials.get("host")
        self.port = self.credentials.get("port")
        self.username = self.credentials.get("user")
        self.password = self.credentials.get("password")
        self.catalog = self.credentials.get("catalog")
        self.schema = self.credentials.get("schema")
        self.verify = self.credentials.get("verify")

        self.connection_string = (
            f"trino://{self.username}@{self.host}:{self.port}/{self.catalog}"
        )
        self.connect_args = {
            "verify": self.verify,
            "auth": BasicAuthentication(self.username, self.password),
            "http_scheme": self.http_scheme,
        }
        self.engine = create_engine(
            self.connection_string, connect_args=self.connect_args, future=True
        )

    @contextmanager
    def get_connection(self):
        """Provide a transactional scope around a series of operations.

        ----
        Examples:

        >>> trino = Trino()
        >>> with trino.get_connection() as connection:
        >>>    trino.run(query1, connection=connection)
        >>>    trino.run(query2, connection=connection)
        """
        connection = self.engine.connect()
        try:
            yield connection
            connection.commit()
        except:
            connection.rollback()
            raise
        finally:
            connection.close()

    def get_tables(self, schema_name: str) -> list[str]:
        query = f"SHOW TABLES FROM {schema_name}"
        with self.get_connection() as connection:
            return list(self.run(query, connection=connection))

    def drop_table(self, table_name: str, schema_name: str = None) -> None:
        fqn = get_fqn(schema_name=schema_name, table_name=table_name)
        query = f"DROP TABLE IF EXISTS {fqn}"

        self.logger.info(f"Dropping table '{fqn}'...")
        with self.get_connection() as connection:
            self.run(query, connection=connection)
        self.logger.info(f"Table '{fqn}' has been successfully dropped.")

    def delete_table(self, table_name: str, schema_name: str = None) -> None:
        fqn = get_fqn(schema_name=schema_name, table_name=table_name)
        query = f"DELETE FROM {fqn}"
        self.logger.info(f"Removing all data from table '{fqn}'...")
        with self.get_connection() as connection:
            self.run(query, connection=connection)
        self.logger.info(f"Data from table '{fqn}' has been successfully removed.")

    def _check_if_table_exists(self, table_name: str, schema_name: str) -> None:
        query = f"""
SELECT *
FROM INFORMATION_SCHEMA.TABLES
WHERE TABLE_SCHEMA = '{schema_name}'
AND TABLE_NAME = '{table_name}'"""
        with self.get_connection() as connection:
            results = list(self.run(query, connection=connection))
        return len(results) > 0

    def get_schemas(self) -> list[str]:
        query = "SHOW SCHEMAS"
        with self.get_connection() as connection:
            return list(self.run(query, connection=connection))

    def _check_if_schema_exists(self, schema_name: str) -> None:
        query = f"SELECT * FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = '{schema_name}'"
        with self.get_connection() as connection:
            results = list(self.run(query, connection=connection))
        return bool(results)

    def drop_schema(self, schema_name: str, cascade: bool = False) -> None:
        if not self._check_if_schema_exists(schema_name):
            return

        if cascade:
            tables = self.get_tables(schema_name)
            [self.drop_table(schema_name=schema_name, table_name=t) for t in tables]

        self.logger.info(f"Dropping schema {schema_name}...")
        with self.get_connection() as connection:
            self.run(f"DROP SCHEMA {schema_name}", connection=connection)
        self.logger.info(f"Schema {schema_name} has been successfully dropped.")

    def create_iceberg_schema(
        self,
        schema_name: str,
        location: str,
        if_exists: Literal["fail", "skip"] = "fail",
    ) -> None:
        exists = self._check_if_schema_exists(schema_name)

        if exists and if_exists == "fail":
            raise ValueError(f"Schema '{schema_name}' already exists.")
        else:
            self.logger.info(f"Schema '{schema_name}' already exists. Skipping...")
            return

        query = f"""
CREATE SCHEMA {schema_name}
WITH (location = '{location}')
        """
        self.logger.info(f"Creating schema '{schema_name}'...")
        with self.get_connection() as connection:
            self.run(query, connection=connection)
        self.logger.info(f"Schema '{schema_name}' has been successfully created.")

    def create_iceberg_table_from_arrow(
        self,
        table: pa.Table,
        table_name: str,
        schema_name: str | None = None,
        location: str | None = None,
        format: Literal["PARQUET", "ORC"] = "PARQUET",
        partition_cols: list[str] | None = None,
        sort_by: list[str] | None = None,
    ) -> None:
        columns = table.schema.names
        types = [self.pyarrow_to_trino_type(str(typ)) for typ in table.schema.types]
        create_table_query = self._create_table_query(
            schema_name=schema_name,
            table_name=table_name,
            columns=columns,
            types=types,
            location=location,
            format=format,
            partition_cols=partition_cols,
            sort_by_cols=sort_by,
        )

        fqn = get_fqn(schema_name=schema_name, table_name=table_name)
        self.logger.info(f"Creating table '{fqn}'...")
        with self.get_connection() as connection:
            self.run(create_table_query, connection=connection)
        self.logger.info(f"Table '{fqn}' has been successfully created.")

    def create_iceberg_table_from_pandas(
        self,
        df: pd.DataFrame,
        table_name: str,
        schema_name: str = None,
        location: str = None,
        format: Literal["PARQUET", "ORC"] = "PARQUET",
        partition_cols: list[str] = None,
        sort_by: list[str] = None,
    ) -> None:
        pa_table = pa.Table.from_pandas(df)
        self.create_iceberg_table_from_arrow(
            table=pa_table,
            schema_name=schema_name,
            table_name=table_name,
            location=location,
            format=format,
            partition_cols=partition_cols,
            sort_by=sort_by,
        )

    def _create_table_query(
        self,
        table_name: str,
        columns: list[str],
        types: list[str],
        schema_name: str = None,
        location: str = None,
        format: Literal["PARQUET", "ORC"] = "PARQUET",
        partition_cols: list[str] = None,
        sort_by_cols: list[str] = None,
    ):
        cols_and_dtypes = ",\n\t".join(
            col + " " + dtype for col, dtype in zip(columns, types)
        )
        fqn = get_fqn(schema_name=schema_name, table_name=table_name)
        with_clause = f"format = '{format}'"

        if partition_cols:
            with_clause += ",\n\tpartitioning = ARRAY" + str(partition_cols)

        if sort_by_cols:
            with_clause += ",\n\tsorted_by = ARRAY" + str(sort_by_cols)

        if location:
            with_clause += f",\n\tlocation = '{location}'"

        query = f"""
CREATE TABLE IF NOT EXISTS {fqn} (
    {cols_and_dtypes}
)
WITH (
    {with_clause}
)"""

        return query

    def run(
        self, sql: str, connection: Connection
    ) -> Generator[tuple, None, None] | None:
        def row_generator(result):
            # Fetch rows in chunks of size `yield_per`.
            # This has to be inside a function due to how Python generators work.
            for partition in result.partitions():
                yield from partition

        self.logger.debug("Executing SQL:\n" + sql)

        try:
            # Execute with server-side cursor of size 5000.
            result = connection.execution_options(yield_per=5000).execute(text(sql))
        except Exception as e:
            raise ValueError(f"Failed executing SQL:\n{sql}") from e

        query_keywords = ["SELECT", "SHOW", "PRAGMA", "WITH"]
        is_query = any(sql.strip().upper().startswith(word) for word in query_keywords)

        if is_query:
            return row_generator(result)

    @staticmethod
    def pyarrow_to_trino_type(pyarrow_type: str) -> str:
        mapping = {
            "string": "VARCHAR",
            "large_string": "VARCHAR",
            "int8": "TINYINT",
            "int16": "SMALLINT",
            "int32": "INTEGER",
            "int64": "BIGINT",
            "float": "REAL",
            "double": "DOUBLE",
            "bool": "BOOLEAN",
            "date32[day]": "DATE",
            "timestamp[ns]": "TIMESTAMP(6)",
            "decimal": "DECIMAL",
            "decimal128": "DECIMAL",
            "decimal256": "DECIMAL",
        }
        precision, scale = None, None
        decimal_match = re.match(r"(\w+)\((\d+), (\d+)\)", pyarrow_type)
        if decimal_match:
            pyarrow_type = decimal_match.group(1)
            precision = int(decimal_match.group(2))
            scale = int(decimal_match.group(3))

        mapped_type = mapping.get(pyarrow_type) or "VARCHAR"

        if precision and scale:
            mapped_type += f"({precision}, {scale})"

        return mapped_type

    def _check_connection(self):
        try:
            with self.get_connection() as connection:
                self.run("select 1", connection=connection)
        except Exception as e:
            raise ValueError(f"Failed to connect to Trino server at {self.host}") from e
